import pytest
import torch
from irtorch.plotter import Plotter
from irtorch.models import BaseIRTModel
from irtorch.estimation_algorithms import MML
from unittest.mock import MagicMock

@pytest.fixture
def mock_model():
    item_categories = [2, 3, 4, 4, 2]
    def model_forward_mock(input_tensor: torch.Tensor):
        return torch.randn(input_tensor.shape[0], 5, 4)

    mock_model = MagicMock(spec = BaseIRTModel, side_effect=model_forward_mock)
    mock_model.latent_variables = 2
    mock_model.mc_correct = None
    mock_model.scale = []
    mock_model.items = 5
    mock_model.item_categories = item_categories
    mock_model.algorithm = MagicMock(spec = MML)
    mock_model.algorithm.training_history = {
        "train_loss": [0.9, 0.8, 0.7],
        "validation_loss": [0.85, 0.75, 0.65]
    }
    return mock_model

@pytest.fixture
def plotter_instance(mock_model):
    return Plotter(mock_model)

def test_plot_training_history_success(plotter_instance: Plotter):
    fig = plotter_instance.plot_training_history()
    assert fig is not None
    assert fig.data is not None
    assert fig.layout.title.text == "Training History"

def test_plot_training_history_not_trained(plotter_instance: Plotter):
    plotter_instance.model.algorithm.training_history = {"train_loss": [], "validation_loss": []}
    with pytest.raises(AttributeError, match="Model has not been trained yet"):
        plotter_instance.plot_training_history()

def test_plot_latent_score_distribution_success(plotter_instance: Plotter):
    scores_to_plot = torch.rand((100, 2))
    fig = plotter_instance.plot_latent_score_distribution(scores_to_plot=scores_to_plot)
    assert fig is not None
    assert fig.data is not None
    assert fig.layout.title.text == None

def test_plot_latent_score_distribution_invalid_dim(plotter_instance: Plotter):
    with pytest.raises(ValueError, match="Can only plot 1 or 2 latent variables."):
        plotter_instance.plot_latent_score_distribution(latent_variables=(1, 2, 3))

def test_plot_item_entropy_success(plotter_instance: Plotter):
    plotter_instance.model.latent_variables = 1
    def probabilities_from_output(output: torch.Tensor) -> torch.Tensor:
        reshaped_output = output.reshape(-1, 4)
        return torch.nn.functional.softmax(reshaped_output, dim=1).reshape_as(output)
    
    plotter_instance.model.probabilities_from_output = MagicMock(side_effect=probabilities_from_output)
    fig = plotter_instance.plot_item_entropy(item=1)
    assert fig is not None
    assert fig.data is not None

def test_plot_item_entropy_invalid_latent_vars(plotter_instance: Plotter):
    with pytest.raises(TypeError, match="Cannot plot more than two latent variables in one plot."):
        plotter_instance.plot_item_entropy(item=1, latent_variables=(1, 2, 3))

def test_plot_item_latent_variable_relationships_success(plotter_instance: Plotter):
    relationships = torch.rand((5, 2))
    fig = plotter_instance.plot_item_latent_variable_relationships(relationships=relationships)
    assert fig is not None
    assert fig.data is not None
    assert fig.layout.title.text == "Relationships: Items vs. latent variables"

def test_plot_item_probabilities_success(plotter_instance: Plotter):
    plotter_instance.model.latent_variables = 1
    plotter_instance.model.item_probabilities = MagicMock(return_value=torch.rand((100, 5, 4)))
    plotter_instance.model.evaluate._min_max_theta_for_integration = MagicMock(return_value=(torch.tensor([-3]), torch.tensor([3])))
    fig = plotter_instance.plot_item_probabilities(item=1)
    assert fig is not None
    assert fig.data is not None
    assert "IRF - Item 1" in fig.layout.title.text

def test_plot_information_success(plotter_instance: Plotter):
    plotter_instance.model.latent_variables = 1
    plotter_instance.model.information = MagicMock(return_value=torch.rand((100,)))
    fig = plotter_instance.plot_information()
    assert fig is not None
    assert fig.data is not None

def test_plot_expected_sum_score_success(plotter_instance: Plotter):
    plotter_instance.model.latent_variables = 1
    plotter_instance.model.expected_scores = MagicMock(return_value=torch.rand((100,)))
    fig = plotter_instance.plot_expected_sum_score()
    assert fig is not None
    assert fig.data is not None
    assert "Expected sum score" in fig.layout.title.text



from .negation_layer import NegationLayer
from .softplus_linear import SoftplusLinear
from .monotone_polynomial import MonotonePolynomialModule
from .rational_quadratic_spline import RationalQuadraticSpline
from .neural_spline_flow import NeuralSplineFlow



import torch

def linear_regression(x, y):
    """
    Performs linear regression.

    Parameters
    -----------
    x : torch.Tensor
        Design matrix of shape (m, n), where m is the number of samples and n is the number of features.
    y : torch.Tensor
        Target vector of shape (m, o), where m is the number of samples and o is the number response variables. If o > 1, this implies a multivariate linear regression.

    Returns
    -----------
    torch.Tensor
        A tensor vector with the bias and the weights of shape (n+1, o).
    """
    # Add a bias term (1) to each sample in x
    bias = torch.cat([torch.ones(x.size(0), 1), x], dim=1)

    # Compute the weights using the normal equation
    w = torch.inverse(bias.t().mm(bias)).mm(bias.t()).mm(y)

    return w

def correlation_matrix(x):
    """
    Compute the covariance matrix from a 2D torch tensor, handling missing values (NaNs).
    
    Parameters
    -----------
    x : torch.Tensor
        A 2D tensor where each column represents a variable and each row represents an observation.
    
    Returns
    -----------
    torch.Tensor
        The covariance matrix of the input tensor, ignoring NaNs.
    """
    if x.dim() != 2:
        raise ValueError("Input must be a 2D tensor")
    mask = ~torch.isnan(x)

    if (~mask).sum() == 0: # No NaNs is more efficient
        x_centered = x - x.mean(dim=0, keepdim=True)
        cov_matrix = x_centered.T @ x_centered / (x.size(0) - 1)
        stddev = x.std(dim=0, unbiased=True)
        corr_matrix = cov_matrix / (stddev.unsqueeze(1) * stddev.unsqueeze(0))
        
        return corr_matrix

    means = torch.nanmean(x, dim=0)
    centered_tensor = x - means
    # Set NaNs back where they originally were in the centered tensor
    centered_tensor[~mask] = 0
    # Calculate the dot product of the centered matrix to get the covariance
    cov_matrix = torch.matmul(centered_tensor.T, centered_tensor)
    
    # Normalization factors: counts of non-NaN pairs
    norm_factors = torch.matmul(mask.T.float(), mask.float())
    # Adjust normalization by dividing by n-1
    norm_factors = torch.where(norm_factors > 1, norm_factors - 1, torch.ones_like(norm_factors))
    cov_matrix = cov_matrix / norm_factors
    
    corr_matrix = torch.ones_like(cov_matrix)
    for i in range(cov_matrix.size(0)):
        for j in range(cov_matrix.size(1)):
            if i == j or i > j:
                continue
            else:
                valid_mask = mask[:, i] & mask[:, j]
                if torch.sum(valid_mask) > 1:
                    std_i = torch.sqrt(torch.sum((x[valid_mask, i] - x[valid_mask, i].mean()) ** 2) / (torch.sum(valid_mask) - 1))
                    std_j = torch.sqrt(torch.sum((x[valid_mask, j] - x[valid_mask, j].mean()) ** 2) / (torch.sum(valid_mask) - 1))
                    corr_matrix[i, j] = cov_matrix[i, j] / (std_i * std_j)
                    corr_matrix[j, i] = corr_matrix[i, j]
                else:
                    corr_matrix[i, j] = 0.0 # Assign a default value if there aren't enough valid pairs
                    corr_matrix[j, i] = 0.0
    
    return corr_matrix

def random_guessing_data(
    item_categories: list[int],
    size: int,
    guessing_probabilities: list[float] = None,
    mc_correct: list[int] = None
):
    """
    Create random test data based on guessing.

    Parameters
    ----------
    item_categories : list[int]
        A list of integers where each integer is the number of possible responses for the corresponding item, excluding missing responses.
    size : int
        The number of rows in the returned tensor.
    guessing_probabilities: list[float], optional
        The guessing probability for each item. The same length as the number of items. Guessing is not supported for polytomously scored items and the probabilities for them will be ignored by setting all responses to 0. (default is None and uses no guessing or, for multiple choice models, 1 over the number of item categories)
    mc_correct : list[int], optional
        Only for multiple choice data with guessing_probabilities supplied. A list of integers where each integer is the correct response for the corresponding item. If None, the data is assumed to be non multiple choice (or dichotomously scored multiple choice with only 0's and 1's). (default is None)

    Returns
    -------
    torch.Tensor
        A 2D tensor with test data. Rows are respondents and columns are items.
    """
    num_items = len(item_categories)
    if (
        guessing_probabilities is not None and \
        isinstance(guessing_probabilities, list) and \
        all(isinstance(item, float) for item in guessing_probabilities) and \
        len(guessing_probabilities) == num_items
    ):
        if not all(0 <= num < 1 for num in guessing_probabilities):
            raise ValueError("The guessing probabilities must be between 0 and 1.")
    else:
        raise ValueError("guessing_probabilities must be list of floats with the same length as the number of items.")


    if guessing_probabilities is None:
        guessing_probabilities = [1.0 / cats for cats in item_categories]

    response_matrix = torch.zeros(size, num_items)

    for item_idx, num_categories in enumerate(item_categories):
        guessing_prob = guessing_probabilities[item_idx]

        if num_categories == 2:
            responses = torch.bernoulli(torch.full((size,), guessing_prob))
        else:
            if mc_correct is not None:
                probs = torch.full((num_categories,), (1 - guessing_prob) / (num_categories - 1))
                
                probs[mc_correct[item_idx]] = guessing_prob

                # For the incorrect answers, we distribute the remaining probability
                incorrect_total_prob = 1.0 - guessing_prob
                probs[probs != guessing_prob] = incorrect_total_prob / (num_categories - 1)

                responses = torch.multinomial(probs.repeat(size, 1), 1).squeeze()
            else:
                responses = torch.zeros(size, dtype=torch.long)

        response_matrix[:, item_idx] = responses

    return response_matrix

def conditional_score_distribution(
    probabilities: torch.Tensor,
    item_categories: list[int],
):
    """
    Compute total score distribution conditional on theta as by (Thissen et. al. 1995).

    Parameters
    ----------
    probabilities : torch.Tensor
        A 3D tensor containing item score probabilities for each theta quadrature point.
        The first dimension represents the quadrature points, the second dimension represents the items and the third dimension represents the item categories.
    item_categories : list[list]
        A list of integers where each integer is the number of possible responses for the corresponding item.

    Returns
    -------
    torch.Tensor
        A 2D torch tensor of total score probabilities, rows being quadrature theta points and columns being total scores. (rows sum to 1)
    """
    no_items = probabilities.shape[1]
    q_points = probabilities.shape[0]
    max_score = sum(item_categories) - no_items

    # Rows are quadrature points, columns are total scores
    nsprobs = torch.zeros(q_points, max_score + 1).to(probabilities.device)
    nsprobs[:, : item_categories[0]] = probabilities[:, 0, :item_categories[0]]

    if no_items > 1:
        for item in range(1, no_items):
            item_max = sum(item_categories[: item + 1]) - len(
                item_categories[: item + 1]
            )
            sprobs = nsprobs.clone()
            nsprobs = torch.zeros_like(nsprobs)
            for j in range(item_max - item_categories[item] + 2):
                nsprobs[:, j : item_categories[item] + j] = (
                    torch.t(sprobs[:, j] * torch.t(probabilities[:, item, :item_categories[item]]))
                    + nsprobs[:, j : item_categories[item] + j]
                )
    return nsprobs


def sum_incorrect_probabilities(
    probabilities: list[torch.Tensor],
    item_responses: list[int],
    mc_correct: list[int],
):
    """
    Sum incorrect score probabilities for multiple choice items. Useful for approximating sum scores.

    Parameters
    ----------
    probabilities : list[torch.Tensor]
        A list of 2D tensors containing item score probabilities for each item.
        Rows are theta quadrature points from the theta density and columns correspond to item responses. (rows sum to 1)
    item_responses : list[int]
        A list of integers where each integer is the number of possible responses for the corresponding item.
    mc_correct : list[int]
        A list of integers where each integer is correct response for the corresponding item.

    Returns
    -------
    torch.Tensor
        A 2D torch tensor correct/incorrect response probabilities
    """
    new_probs = torch.zeros(probabilities.shape[0], probabilities.shape[1], 2)
    for item in range(0, len(item_responses)):
        item_score_0_probs = torch.cat(
            (
                probabilities[:, item, :mc_correct[item]],
                probabilities[:, item, mc_correct[item]+1:]
            ),
            dim=1,
        ).sum(dim=1)
        new_probs[:, item, :] = torch.cat(
            (
                item_score_0_probs.unsqueeze(1),
                probabilities[:, item, mc_correct[item]].unsqueeze(1),
            ),
            dim=1,
        )
    return new_probs


def entropy(probabilities: torch.Tensor, log_base: float = 2.0):
    """
    Calculate the entropy of a set of probabilities.

    The entropy is a measure of the uncertainty or randomness of a set of probabilities. It is calculated as the sum of the product of each probability and its surprisal (negative log probability).

    Parameters
    -------------
    probabilities: torch.Tensor
        A tensor of probabilities. The last tensor dimension should represent a discrete probability distribution and should sum to 1.
    log_base: float, optional
        The base of the logarithm used in the calculation. (default is 2, which gives entropy in bits)

    Returns
    -------------
    torch.Tensor
        The entropies of the probabilities.
    """
    if not torch.allclose(probabilities.sum(dim=-1), torch.tensor(1.0)):
        raise RuntimeError("The probabilities of the last dimension must sum to 1.")
    # Compute log probabilities safely by adding a small epsilon to avoid log(0) and maintain gradient flow
    epsilon = 1e-10
    surprisal = -torch.log(probabilities + epsilon) / torch.log(torch.tensor(log_base))
    surprisal[surprisal.isinf()] = 0 # if surprisal is Inf, set to finite value to get correct entropy instead of NaN
    entropies = (probabilities * surprisal).sum(dim=-1)
    return entropies

def joint_entropy_matrix(data: torch.Tensor, log_base: float = 2.0):
    """
    Calculate the matrix of joint entropies from test data.

    Parameters
    -------------
    data: torch.Tensor
        A tensor with test data.
    log_base: float, optional
        The base of the logarithm used in the calculation. (default is 2, which gives entropy in bits)

    Returns
    -------------
    torch.Tensor
        The joint entropy matrix.
    """
    if data.dim() != 2:
        raise ValueError("Input must be a 2D tensor")
    
    mask = ~torch.isnan(data)
    int_data = data.int()
    max_combos = ((int_data.max() + 1) ** 2).item()
    proportions = torch.zeros(int_data.shape[1], int_data.shape[1], max_combos)
    for i in range(int_data.shape[1]):
        for j in range(int_data.shape[1]):
            if i > j:
                continue
            else:
                valid_mask = mask[:, i] & mask[:, j]
                if torch.sum(valid_mask) > 1:
                    relevant_data = int_data[valid_mask, :]
                    unique_combinations, counts = torch.unique(
                        relevant_data[:, [i, j]], dim=0, return_counts=True
                    )
                    proportions[i, j, :unique_combinations.size(0)] = counts.float() / counts.sum()
                    proportions[j, i, :unique_combinations.size(0)] = proportions[i, j, :unique_combinations.size(0)]

    return entropy(proportions, log_base)

def is_jupyter():
    try:
        from IPython import get_ipython
        if "IPKernelApp" not in get_ipython().config:
            raise ImportError("Not in IPython")
        return True
    except Exception:
        return False

def dynamic_print(string_to_print):
    """
    Dynamically update terminal printout.

    Parameters
    ----------
    *args : str
        Strings to print

    """
    formatted_string = f"\r{string_to_print} " # small space after to make it look better in terminal
    print(formatted_string, end="", flush=True)

def one_hot_encode_test_data(
    data: torch.Tensor, item_categories: list
):
    """
    One-hot encodes test data for each item based on the number of response categories for that item. Missing item responses need to be coded as -1 or nan.

    Parameters
    ----------
    data : torch.Tensor
        A 2D tensor where each row represents one respondent and each column represents an item.
        The values should be the scores achieved by the respondents on the items, starting from 0.
        Missing item responses need to be coded as -1 or 'nan'.
    item_categories : list
        A list of integers where each integer is the number of response categories for the corresponding item.

    Returns
    -------
    torch.Tensor
        A 2D tensor where each group of columns corresponds to the one-hot encoded scores for one of the items.
        The number of columns in each group is equal to the maximum possible score plus one for that item.

    Notes
    -----
    The input data tensor should contain integer values representing the scores. If it contains non-integer values,
    they will be rounded to the nearest integer.
    """
    if data.dim() != 2:
        raise ValueError("Input data must be a 2D tensor.")
    if data.shape[1] != len(item_categories):
        raise ValueError(
            "The number of columns in the data tensor must match the length of the item_categories list."
        )
    if data.isnan().any():
        data[data.isnan()] = -1

    one_hot_list = []
    if data.dtype != torch.long:
        data = data.round().long()

    for i in range(data.shape[1]):
        one_hot = torch.zeros((data.shape[0], item_categories[i]), dtype=torch.long).to(
            data.device
        )
        # Fill in the appropriate column with ones based on the scores
        # Only for those rows where the score is not -1
        valid_rows = data[:, i] != -1
        one_hot[valid_rows, data[valid_rows, i]] = 1
        # Append the one-hot encoded tensor to the list
        one_hot_list.append(one_hot)

    # Concatenate the one-hot encoded columns back into a single tensor
    return torch.cat(one_hot_list, dim=1).float()

def get_missing_mask(data: torch.Tensor) -> torch.Tensor:
    """
    Get a mask for missing values in the data.

    Parameters
    ----------
    data : torch.Tensor
        The data tensor.

    Returns
    -------
    torch.Tensor
        A tensor of the same shape as the input data tensor, with 1s where the data is missing and 0s where it is not.
    """
    return (data == -1) | data.isnan()

@torch.no_grad()
def impute_missing_internal(
    data: torch.Tensor,
    method: str = "zero",
    mc_correct: list[int] = None,
    item_categories: list[int] = None,
) -> torch.Tensor:
    """
    Impute missing values. Separate from the external version to not rely on a fitted model. 

    Parameters
    ----------
    data : torch.Tensor
        A 2D tensor where each row is a response vector and each column is an item.
    method : str, optional
        The imputation method to use. Options are 'zero', 'mean'. (default is 'zero')

        - 'zero': Impute missing values with 0.
        - 'mean': Impute missing values with the item means.
    mc_correct : list[int], optional
        Only for method='random_incorrect'. A list of integers where each integer is the correct response for the corresponding item. If None, the data is assumed to be non multiple choice (or dichotomously scored multiple choice with only 0's and 1's). (default is None)
    item_categories : list[int], optional
        Only for method='random_incorrect'. A list of integers where each integer is the number of possible responses for the corresponding item. If None, the number of possible responses is calculated from the data. (default is None)
    """
    imputed_data = data.clone()

    if (imputed_data == -1).any():
        imputed_data[imputed_data == -1] = torch.nan

    if method == "zero":
        imputed_data = torch.where(torch.isnan(imputed_data), torch.tensor(0.0), imputed_data)
    elif method == "mean":
        means = imputed_data.nanmean(dim=0)
        mask = torch.isnan(imputed_data)
        imputed_data[mask] = means.repeat(data.shape[0], 1)[mask]
    elif method == "random incorrect":
        if mc_correct is None:
            raise ValueError("mc_correct must be provided when using random_incorrect imputation without a model")
        if item_categories is None:
            item_categories = (torch.where(~imputed_data.isnan(), data, torch.tensor(float('-inf'))).max(dim=0).values + 1).int().tolist()

        for col in range(imputed_data.shape[1]):
            # Get the incorrect non-missing responses from the column
            incorrect_responses = torch.arange(0, item_categories[col], device=imputed_data.device).float()
            incorrect_responses = incorrect_responses[incorrect_responses != mc_correct[col]]
            # Find the indices of missing values in the column
            missing_indices = (imputed_data[:, col].isnan()).squeeze()
            # randomly sample from the incorrect responses and replace missing
            imputed_data[missing_indices, col] = incorrect_responses[torch.randint(0, incorrect_responses.size(0), (missing_indices.sum(),))]
    else:
        raise ValueError(
            f"{method} imputation is not implmented"
        )
        

    return imputed_data

def sum_score(data: torch.Tensor, mc_correct: list[int] = None):
    """
    Calculate the sum score for each respondent.

    Parameters
    ----------
    data : torch.Tensor
        A 2D tensor where each row is a response vector and each column is an item.
    mc_correct : list[int], optional
        Only for multiple choice data. A list of integers where each integer is the correct response for the corresponding item. If None, the data is assumed to be non multiple choice and the rows are summed. (default is None)

    Returns
    -------
    torch.Tensor
        A tensor of sum scores for each respondent.
    """
    if mc_correct is None:
        return data.sum(dim=1)
    else:
        return (data == torch.tensor(mc_correct, dtype=data.dtype)).int().sum(dim=1).float()



import logging
import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.distributions import Distribution, Normal
from irtorch.rescale import Scale
from irtorch.torch_modules import NeuralSplineFlow
from irtorch.torch_modules import RationalQuadraticSpline

logger = logging.getLogger("irtorch")

class Flow(Scale):
    """
    Normalizing flow transformation of IRT theta scales using rational quadratic splines as per :cite:t:`Durkan2019`.
    Supports gradient computation and the transformation is invertible.
    
    Parameters
    ----------
    latent_variables : int
        The number of latent variables.
    
    Examples
    --------
    >>> import irtorch
    >>> from irtorch.models import GradedResponse
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.rescale import Flow
    >>> data = irtorch.load_dataset.swedish_national_mathematics_1()
    >>> model = GradedResponse(data)
    >>> model.fit(train_data=data, algorithm=AE())
    >>> thetas = model.latent_scores(data)
    >>> # Initalize and fit the flow scale transformation. Supply it to the model.
    >>> flow = Flow(1)
    >>> flow.fit(thetas)
    >>> model.add_scale_tranformation(flow)
    >>> # Estimate thetas on the transformed scale
    >>> rescaled_thetas = model.latent_scores(data)
    >>> # Or alternatively by directly converting the old ones
    >>> rescaled_thetas = model.transform_theta(thetas)
    >>> # Plot the differences
    >>> model.plot.plot_latent_score_distribution(thetas).show()
    >>> model.plot.plot_latent_score_distribution(rescaled_thetas).show()
    >>> # Put the thetas back to the original scale
    >>> original_thetas = model.inverse_transform_theta(rescaled_thetas)
    >>> # Plot an item on the flow transformed scale
    >>> model.plot.plot_item_probabilities(1).show()
    """
    def __init__(self, latent_variables: int):
        super().__init__(invertible=True)
        self.latent_variables = latent_variables
        self.theta_means = torch.zeros(latent_variables)
        self.theta_stds = torch.ones(latent_variables)
        self.flow = None

    def fit(
        self,
        theta:torch.Tensor,
        transformation: RationalQuadraticSpline = None,
        distribution: Distribution = None,
        batch_size: int = None,
        learning_rate: float = 0.01,
        learning_rate_updates_before_stopping: int = 2,
        evaluation_interval_size: int = 50,
        max_epochs: int = 1500,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        **kwargs
    ):
        """
        Fits the normalizing flow to the data. Typically used from within an IRT model instance. Use batch_size if the data is too large to fit in memory.
        
        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores of the population. Usually the training data. Each column represents one latent variable.
        transformation : RationalQuadraticSpline, optional
            The transformation to apply to the data.
        distribution : Distribution, optional
            The distribution to apply to the latent variables. If None, a standard normal distribution is used.
        batch_size : int, optional
            The batch size for the data loader. Default is None and uses no batches.
        learning_rate : float, optional
            The learning rate for the optimizer. Default is 0.01.
        learning_rate_updates_before_stopping, optional
            The number of learning rate updates before stopping the training. Default is 2.
        evaluation_interval_size: int, optional
            The number of iterations between each model evaluation during training. (default is 50)
        max_epochs : int, optional
            The maximum number of epochs to train the flow. Default is 1500.
        device : str, optional
            The device to use for the computation. Default is "cuda" if available, otherwise "cpu".
        **kwargs
            Additional keyword arguments for :class:`irtorch.torch_modules.RationalQuadraticSpline` constructor.
        """
        if transformation is None:
            spline_params = {
                'lower_input_bound': -5.5,
                'upper_input_bound': 5.5,
                'lower_output_bound': -3.0,
                'upper_output_bound': 3.0,
                'num_bins': 300,
            }
            spline_params.update(kwargs)
            transformation = RationalQuadraticSpline(
                self.latent_variables,
                **spline_params
            )

        if distribution is None:
            distribution = Normal(
                torch.zeros(1).to(device),
                torch.ones(1).to(device)
            )
        if len(distribution.batch_shape) > 0 and distribution.batch_shape[0] not in [1, self.latent_variables]:
            raise ValueError("The distribution batch shape should match the number of latent variables or be 0.")

        if batch_size is None:
            batch_size = theta.shape[0]

        # standardize the data and store the mean and std for later use
        self.theta_means = theta.mean(dim=0)
        self.theta_stds = theta.std(dim=0)
        theta = (theta - self.theta_means) / self.theta_stds

        self.flow = NeuralSplineFlow(transformation, distribution).to(device)
        optimizer = torch.optim.Adam(self.flow.parameters(), lr=learning_rate, amsgrad=True)
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, mode="min", factor=0.6, patience=1
        )
        loader = DataLoader(TensorDataset(theta.to(device)), batch_size=batch_size, shuffle=True)

        lr_update_count = 0
        total_iterations = 0
        prev_losses = []
        for epoch in range(max_epochs):
            epoch_losses = []
            for batch in loader:
                loss = -self.flow.log_prob(batch[0]).mean()
                if torch.isnan(loss):
                    logger.warning("Loss is NaN. Stopping training. Try increasing batch size or lowering the learning rate.")
                    break

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                total_iterations += 1
                prev_losses.append(loss.item())
                epoch_losses.append(loss.item())
                if len(prev_losses) > evaluation_interval_size:
                    prev_losses.pop(0)

                # Update the learning rate scheduler every self.evaluation_interval_size iterations if no improvement
                if (total_iterations) % evaluation_interval_size == 0 and total_iterations != 1:
                    mean_loss = np.mean(prev_losses)
                    scheduler.step(mean_loss)
                    logger.info("Iteration %s: Mean loss: %s", total_iterations, mean_loss.item())

                # Check if the learning rate has been updated
                if learning_rate != optimizer.param_groups[0]['lr']:
                    learning_rate = optimizer.param_groups[0]['lr']
                    lr_update_count += 1

                if lr_update_count >= learning_rate_updates_before_stopping:
                    logger.info("Stopping training after %s learning rate updates.", learning_rate_updates_before_stopping)
                    break
            
            if torch.isnan(loss):
                break
            if lr_update_count >= learning_rate_updates_before_stopping:
                break

            logger.info("Epoch %s: Loss: %s", epoch, np.mean(epoch_losses))

        self.flow.to("cpu")
        self.flow.eval()

    def _flow_exists(self):
        if self.flow is None:
            raise ValueError("The flow has not been initialized. Please fit the flow first using the fit method.")

    def transform(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Transforms the input theta scores into the new scale.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.
        """
        self._flow_exists()
        theta = (theta - self.theta_means) / self.theta_stds
        return self.flow(theta)

    def inverse(self, transformed_theta: torch.Tensor) -> torch.Tensor:
        """
        Puts the scores back to the original theta scale.

        Parameters
        ----------
        transformed_theta : torch.Tensor
            A 2D tensor containing transformed theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A 2D tensor containing theta scores on the the original scale.
        """
        self._flow_exists()
        theta = self.flow.inverse(transformed_theta)
        # destandardize
        return theta * self.theta_stds + self.theta_means

    def jacobian(
        self,
        theta: torch.Tensor
    ) -> torch.Tensor:
        r"""
        Computes the Jacobian of scale scores for each :math:`j` with respect to the input theta scores.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A tensor with the Jacobian for each input row. Dimensions are (theta rows, latent variables, latent variables) where the last two are the jacobians.
        """
        self._flow_exists()
        theta_scores = theta.clone()
        theta_scores.detach_().requires_grad_(True)
        standardized_theta_scores = (theta_scores - self.theta_means) / self.theta_stds
        transformed_thetas = self.flow(standardized_theta_scores)
        transformed_thetas.sum().backward()
        jacobians = torch.diag_embed(theta_scores.grad) # Since each transformation is only dependent on one theta score
        return jacobians



import torch
from irtorch.rescale import Scale

class RankCDF(Scale):
    """
    Rank-based inverse CDF transformation of IRT theta scales.
    
    For each latent variable, finds the rank of each theta score in the population.
    For new data, finds the closest matching population ranks and uses the inverse CDF to 
    find the equivalents scores in the chosen distribution(s).

    Note that while this method is fast, it is heavily reliant on the input theta scores covering the entire range of the distribution(s).
    It is also not invertible, does not support gradient computation and the transformation is not unique one-to-one.
    
    Parameters
    ----------
    theta : torch.Tensor
        A large tensor of theta scores representing the population.
    distributions : list[torch.distributions.Distribution], optional
        The distributions to use for the transformation of each latent variable. If None, normal distributions are used.
    
    Examples
    --------
    >>> import irtorch
    >>> from irtorch.models import GradedResponse
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.rescale import RankCDF
    >>> data = irtorch.load_dataset.swedish_national_mathematics_1()
    >>> model = GradedResponse(data)
    >>> model.fit(train_data=data, algorithm=AE())
    >>> thetas = model.latent_scores(data)
    >>> # Create and RankCDF instancce and supply it to the model.
    >>> model.add_scale_tranformation(RankCDF(thetas))
    >>> # Estimate thetas on the transformed scale
    >>> rescaled_thetas = model.latent_scores(data)
    >>> # Or alternatively by directly converting the old ones
    >>> rescaled_thetas = model.transform_theta(thetas)
    >>> # Plot the differences
    >>> model.plot.plot_latent_score_distribution(thetas).show()
    >>> model.plot.plot_latent_score_distribution(rescaled_thetas).show()
    >>> # Plot an item on the flow transformed scale
    >>> model.plot.plot_item_probabilities(1).show()
    """
    def __init__(self, theta: torch.Tensor, distributions: list[torch.distributions.Distribution] = None):
        super().__init__(invertible=False)
        self.n_samples, latent_variables = theta.shape
        if distributions is None:
            self.distributions = []
            for _ in range(theta.shape[1]):
                self.distributions.append(torch.distributions.Normal(0, 1))
        else:
            if theta.shape[1] != len(distributions) and len(distributions) != 1:
                raise ValueError("The number of distributions must be one or equal to the number of latent variables.")
            if len(distributions) == 1:
                self.distributions = []
                for _ in range(theta.shape[1]):
                    self.distributions.append(distributions[0])
            else:
                self.distributions = distributions

        self.unique_vals_list = []
        self.avg_ranks_list = []

        for feature in range(latent_variables):
            sorted_data = torch.sort(theta[:, feature]).values
            # Generate rank indices for the sorted data
            rank = torch.arange(1, self.n_samples + 1, dtype=torch.float32)
            # Handle ties: Compute unique values and assign average rank for ties
            unique_vals, inverse_indices = torch.unique(sorted_data, return_inverse=True, sorted=True)
            # Sum ranks for the same unique values (i.e., tie groups) and count the number of occurrences per group
            rank_sum = torch.zeros_like(unique_vals)
            rank_count = torch.zeros_like(unique_vals)
            rank_sum = rank_sum.scatter_add(0, inverse_indices, rank)
            rank_count = rank_count.scatter_add(0, inverse_indices, torch.ones_like(rank))
            # Compute the average rank for each unique value
            avg_ranks = rank_sum / rank_count
            self.avg_ranks_list.append(avg_ranks)
            self.unique_vals_list.append(unique_vals)

    def transform(self, theta):
        """
        Transforms the input theta scores into the new scale.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing transformed theta scores. Each column represents one latent variable.
        """
        n_samples, latent_variables = theta.shape
        transformed_data = torch.zeros_like(theta, dtype=torch.float32)

        for latent_variable in range(latent_variables):
            unique_vals = self.unique_vals_list[latent_variable]
            avg_ranks = self.avg_ranks_list[latent_variable]

            # For each theta, find the corresponding rank from the original data
            new_ranks = torch.zeros(n_samples)
            for i in range(n_samples):
                # Find the closest matching unique value
                closest_idx = torch.abs(unique_vals - theta[i, latent_variable]).argmin()
                new_ranks[i] = avg_ranks[closest_idx]

            # Normalize the ranks to (0, 1) and apply the inverse CDF
            rank_normalized = new_ranks / (self.n_samples + 1)
            transformed_data[:, latent_variable] = \
                self.distributions[latent_variable].icdf(rank_normalized).clone().detach()

        return transformed_data

    def inverse(self, transformed_theta):
        raise NotImplementedError("RankCDF is not invertible.")

    def jacobian(
        self,
        theta: torch.Tensor
    ) -> torch.Tensor:
        r"""
        Computes the gradients of scale scores with respect to the input theta scores.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A torch tensor with the gradients for each theta score. Dimensions are (theta rows, latent variables, latent variables) where the last two are the jacobians.
        """
        raise NotImplementedError("Gradients are not available for RankCDF scale transformations.")



import pytest
from unittest.mock import patch, MagicMock
import torch
from irtorch.estimation_algorithms.vae import VAE, VariationalEncoder
from irtorch.models import MonotoneNN, BaseIRTModel


# The @pytest.fixture decorator is used to create fixture methods.
# This method is called once per test method that uses it, and its return value is passed to the test method as an argument.
# pytest.fixture with params creates two fixtures, one for the CPU and one for the GPU.
# The ids parameter is used to give the tests meaningful names
class TestVAE:
    @pytest.fixture()
    def irt_model(self, latent_variables, item_categories):
        model = MonotoneNN(
            latent_variables = latent_variables,
            item_categories = item_categories,
            hidden_dim = [3]
        )
        return model

    @pytest.fixture()
    def algorithm(self, irt_model: BaseIRTModel, data_loaders, latent_variables, item_categories):
        algorithm = VAE()
        algorithm.imputation_method = "zero"
        algorithm.one_hot_encoded = False
        # Set DataLoaders from the fixture
        algorithm.data_loader = data_loaders

        algorithm.encoder = VariationalEncoder(
            input_dim=len(item_categories),
            latent_variables=latent_variables,
            hidden_dim=[2 * sum(irt_model.item_categories)]
        )
        return algorithm

    @pytest.fixture()
    def irt_model_small(self, latent_variables, item_categories_small):
        model = MonotoneNN(
            latent_variables = latent_variables,
            item_categories = item_categories_small,
            hidden_dim = [3]
        )
        return model

    @pytest.fixture()
    def algorithm_small_data(self, irt_model: BaseIRTModel, data_loaders_small, latent_variables, item_categories_small):
        algorithm = VAE()
        algorithm.imputation_method = "zero"
        algorithm.one_hot_encoded = False
        # Set DataLoaders from the fixture
        algorithm.data_loader = data_loaders_small

        algorithm.encoder = VariationalEncoder(
            input_dim=len(item_categories_small),
            latent_variables=latent_variables,
            hidden_dim=[2 * sum(irt_model.item_categories)]
        )
        return algorithm

    def test_theta_scores(self, algorithm: VAE, irt_model: BaseIRTModel, test_data):
        output = algorithm.theta_scores(test_data)
        assert output.shape == (120, irt_model.latent_variables)

    @pytest.mark.parametrize("iw_samples", [1, 3])
    def test__loss_function(
        self,
        algorithm_small_data: VAE,
        irt_model_small: BaseIRTModel,
        test_data,
        iw_samples,
        latent_variables,
    ):
        algorithm_small_data.iw_samples = iw_samples
        logits = torch.tensor(
            [
                [0.8221, -0.3415, -torch.inf, 0.3751, -0.3118, 0.0909],
                [-0.6602, 0.0387, -torch.inf, -0.4591, 0.3182, 0.0278],
            ],
        ).repeat(iw_samples, 1)
        means = torch.tensor([[0], [1]]).repeat(
            iw_samples, latent_variables
        )
        logvars = torch.tensor([[0], [0.5]]).repeat(
            iw_samples, latent_variables
        )
        std = torch.exp(0.5 * logvars)
        theta_samples = means + torch.randn_like(std) * std
        loss = algorithm_small_data._loss_function(
            irt_model_small, test_data[0:2, 0:2], torch.zeros_like(test_data[0:2, 0:2]).bool(), logits, theta_samples, means, logvars
        )
        assert loss > 0

    def test_fit(self, algorithm: VAE, irt_model: BaseIRTModel, test_data):
        # Mock the inner functions that would be called during training
        with patch.object(
            algorithm, "_train_step", return_value=torch.tensor(0.5)
        ) as mocked_train_step:
            algorithm.fit(
                model=irt_model,
                train_data=test_data[0:100],
                max_epochs=5
            )

            # Check if inner functions are called
            assert mocked_train_step.call_count == 5

    def test_latent_credible_interval(self, algorithm: VAE):
            input_data = torch.randn(10, algorithm.encoder.input_dim)

            def encoder_mock(input_data: torch.Tensor) -> torch.Tensor:
                means = (torch.arange(input_data.size(0)) - (input_data.size(0)/2)) / input_data.size(0)
                means = means.view(-1, 1).repeat(input_data.size(0), 2).float()
                variances = (torch.arange(input_data.size(0)) + 1) / input_data.size(0)
                variances = variances.view(-1, 1).repeat(input_data.size(0), 2).float()
                return means, variances.log()

            algorithm.encoder = MagicMock(side_effect=encoder_mock)

            lower, mean, upper = algorithm.latent_credible_interval(input_data, alpha=0.05)

            expected_mean = (torch.arange(input_data.size(0)) - (input_data.size(0)/2)) / input_data.size(0)
            expected_mean = expected_mean.view(-1, 1).repeat(input_data.size(0), 2).float()
            assert torch.all(mean == expected_mean)

            assert lower.shape == mean.shape == upper.shape == (100, 2)
            assert torch.all(lower <= mean)  # Lower bound should be less than or equal to the mean
            assert torch.all(mean <= upper)  # Mean should be less than or equal to the upper bound

    def test_latent_mean_se(self, algorithm: VAE):
        input_data = torch.randn(10, algorithm.encoder.input_dim)

        def encoder_mock(input_data: torch.Tensor) -> torch.Tensor:
            means = (torch.arange(input_data.size(0)) - (input_data.size(0)/2)) / input_data.size(0)
            means = means.view(-1, 1).repeat(input_data.size(0), 2).float()
            variances = (torch.arange(input_data.size(0)) + 1) / input_data.size(0)
            variances = variances.view(-1, 1).repeat(input_data.size(0), 2).float()
            return means, variances.log()

        algorithm.encoder = MagicMock(side_effect=encoder_mock)

        mean, se = algorithm.latent_mean_se(input_data)

        expected_mean = (torch.arange(input_data.size(0)) - (input_data.size(0)/2)) / input_data.size(0)
        expected_mean = expected_mean.view(-1, 1).repeat(input_data.size(0), 2).float()
        assert torch.all(mean == expected_mean)

        assert mean.shape == se.shape == (100, 2)
        assert torch.all(se > 0)



import logging
import random
import copy
from itertools import product
import torch
import torch.multiprocessing as mp
import pandas as pd
import numpy as np
from irtorch.models import BaseIRTModel

__all__ = ["cross_validation", "gauss_hermite", "get_item_categories", "impute_missing", "fit_multiple_models_cpu", "split_data", "set_seed", "imw"]

logger = logging.getLogger("irtorch")

def cross_validation(
    model: BaseIRTModel,
    data: torch.Tensor,
    folds: int,
    params_grid: dict,
    theta_estimation: str = "ML",
    device: str = "cuda" if torch.cuda.is_available() else "cpu",
    cores_to_use: int = None,
    **kwargs
) -> pd.DataFrame:
    """
    Perform cross-validation on the given model and data. Uses log-likelihood for model evaluation. Note that for running on the CPU on windows, `if __name__ == '__main__':` needs to be added to the main script before calling this function, see examples.

    Parameters
    ----------
    model : IRT
        The irt model to train. Note that this should be an untrained model.
    data : torch.Tensor
        The data to use for cross-validation. The data is randomly shuffled before splitting into folds.
    folds : int
        The number of folds to use for cross-validation.
    params_grid : dict
        The hyperparameters to use for cross-validation. All need to be arguments for the model fit method.
    theta_estimation : str, optional
        Method used to obtain the theta scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'ML')
    device : str, optional
        The device to use for training. Can be 'cpu' for CPU or 'cuda' for GPU (if available). The default is 'cuda' if a GPU is available, otherwise 'cpu'.
    **kwargs
        Additional keyword arguments to pass to the model fit method.

    Returns
    -------
    list
        A list of dictionaries containing the hyperparameters and the corresponding cross-validation scores.

    Examples
    --------
    This example demonstrates how to use cross_validation() function with the Swedish National Mathematics dataset.

    First, we import necessary modules, load the data and split it into a training and testing set:

    >>> from irtorch.models import MonotoneNN
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.load_dataset import swedish_national_mathematics_2
    >>> from irtorch.utils import split_data, cross_validation
    >>> data_math = swedish_national_mathematics_2()
    >>> train_data, test_data = split_data(data_math, 0.8)

    Next, we initialize the IRT model:

    >>> model = MonotoneNN(data=data_math)

    We then set up a grid of parameters for cross-validation:

    >>> params_grid = {
    ...     'learning_rate': [0.05, 0.1],
    ...     'batch_size': [64, 128],
    ... }

    Finally, we perform cross-validation to find a good set of parameters:

    >>> if __name__ == '__main__':
    ...     result = cross_validation(model, data=train_data, folds=5, params_grid=params_grid, theta_estimation='NN', device='cpu', algorithm = AE())
    """
    if device == "cuda" and not torch.cuda.is_available():
        raise ValueError("CUDA is not available on this machine, use device = 'cpu'.")

    # randomly scrable the data
    data = data[torch.randperm(data.shape[0])]
    data_folds = torch.chunk(data, folds, dim=0)

    param_combinations = list(product(*params_grid.values()))
    param_comb_names = list(params_grid.keys())
    param_dicts = [
        {name: value for name, value in zip(param_comb_names, combination)}
        for combination in param_combinations
    ]
    # Add additional kwargs to each parameter dictionary
    for param_dict in param_dicts:
        param_dict.update(kwargs)

    logger.info(f"Performing cross-validation with {len(param_dicts)} parameter combinations")

    # Prepare arguments for multiprocessing
    jobs = []
    for params in param_dicts:
        for fold in range(folds):
            train_data = torch.cat([data_folds[i] for i in range(folds) if i != fold])
            validation_data = data_folds[fold]
            jobs.append((copy.deepcopy(model), train_data, validation_data, params, theta_estimation, device))

    if device == "cpu":
        if cores_to_use is None:
            cores_to_use = min(mp.cpu_count(), len(jobs))
        print(f"Using {cores_to_use} cores")
        original_threads = torch.get_num_threads()
        try:
            # Ensure each subprocess uses a single thread for PyTorch.
            torch.set_num_threads(1)
            with mp.Pool(cores_to_use) as pool:
                results = pool.starmap(_cv_fold, jobs)
        finally:
            torch.set_num_threads(original_threads)

    elif device == "cuda":
        results = []
        for job in jobs:
            results.append(_cv_fold(*job))

    # average over folds
    results = pd.DataFrame(results)
    results.drop(list(kwargs.keys()), axis=1, inplace=True)
    results = results.groupby(param_comb_names).mean().reset_index()
    return results

def _cv_fold(irt_model : BaseIRTModel, train_data, validation_data, params, theta_estimation, device):
    if device == "cpu":
        torch.set_num_threads(1) # One thread per core, to avoid overloading the CPU

    irt_model.fit(train_data, device=device, **params)
    log_likelihood = irt_model.evaluate.log_likelihood(validation_data, theta_estimation = theta_estimation, reduction="sum").item()

    return {**params, "log_likelihood": log_likelihood}


def gauss_hermite(n, mean, covariance):
    r"""
    Calculate the Gauss-Hermite quadrature points and weights for a multivariate normal distribution.

    Parameters
    ----------
    n : int
        Number of quadrature points.
    mean : torch.Tensor
        Mean of the distribution.
    covariance : torch.Tensor
        Covariance matrix of the distribution.

    Returns
    -------
    x : torch.Tensor
        Quadrature points.
    w : torch.Tensor
        Quadrature weights.

    Notes
    -----
    To integrate a function against a multivariate Gaussian distribution, one can employ Gauss-Hermite quadrature with an appropriate change of variables. Beginning with the integral:

    .. math::
        \int \frac{1}{\sqrt{\operatorname{det}(2 \pi \mathbf{\Sigma})}} e^{-\frac{1}{2}(\mathbf{y}-\boldsymbol{\mu})^{\top} \mathbf{\Sigma}^{-1}(\mathbf{y}-\boldsymbol{\mu})} f(\mathbf{y}) d \mathbf{y}

    We use the transformation :math:`\mathbf{x}=\frac{1}{\sqrt{2}} \mathbf{L}^{-1}(\mathbf{y}-\boldsymbol{\mu})`, where :math:`\boldsymbol{\Sigma}=\mathbf{L} \mathbf{L}^{\top}`, leading to:

    .. math::
        \int \frac{1}{\sqrt{\operatorname{det}(2 \pi \mathbf{\Sigma})}} e^{\mathbf{x}^{\top} \mathbf{x}} f(\sqrt{2} \mathbf{L} \mathbf{x}+\boldsymbol{\mu}) \operatorname{det}(\sqrt{2} \mathbf{L}) d \mathbf{x}=\int \pi^{-\frac{N}{2}} e^{-\mathbf{x}^{\top} \mathbf{x}} f(\sqrt{2} \mathbf{L} \mathbf{x}+\boldsymbol{\mu}) d \mathbf{x}

    For an :math:`N`-dimensional vector :math:`\boldsymbol{x}`, the integral can be decomposed into :math:`N` nested Gauss-Hermite integrals, since the inner product in the exponent, :math:`\exp \left(\sum_{n=1}^N x_n^2\right)`, can be represented as a product.

    Examples
    --------
    Computing mean and variance of a multivariate normal distribution using 8 Gauss-Hermite quadrature points:

    >>> import torch
    >>> from irtorch.utils import gauss_hermite
    >>> mu = torch.tensor([1, 0])
    >>> cov = torch.tensor([[1.3, -0.213], [-0.213, 1.2]])
    >>> points, weights = gauss_hermite(8, mu, cov)
    >>> mean = torch.sum(weights[:, None] * points, dim=0)
    >>> variance = torch.sum(weights[:, None] * (points - mu)**2, dim=0)
    >>> print(f"Mean: {mean}")
    >>> print(f"Variance: {variance}")
    """
    mvn_dim = len(mean) #dimesions of the multivariate normal distribution
    x, w = np.polynomial.hermite.hermgauss(n)
    x = torch.tensor(x, dtype=torch.float64)
    w = torch.tensor(w, dtype=torch.float64)
    const = torch.pi**(-0.5 * mvn_dim)
    xn = torch.tensor(list(product(*(x,)*mvn_dim)), dtype=torch.float64)
    wn = torch.prod(torch.tensor(list(product(*(w,)*mvn_dim)), dtype=torch.float64), dim=1)
    chol_decomp = torch.linalg.cholesky(covariance).to(torch.float64)
    # Transformation of the quadrature points
    # See change of variables in the docstring
    yn = 2.0**0.5 * (chol_decomp @ xn.T).T + mean[None, :]
    return yn, wn * const

def get_item_categories(data: torch.Tensor):
    """
    Get the number of possible responses for each item in the data.

    Parameters
    -----------
    data : torch.Tensor
        A 2D tensor where each row represents one respondent and each column represents an item.
        The values should be the scores/possible responses on the items, starting from 0.
        Missing item responses need to be coded as -1 or 'nan'.

    Returns
    ----------
    list
        A list of integers where each integer is the number of possible responses for the corresponding item.
    """
    return [int(data[~data.isnan().any(dim=1)][:, col].max()) + 1 for col in range(data.shape[1])]

@torch.no_grad()
def impute_missing(
    data: torch.Tensor,
    method: str = "zero",
    model: BaseIRTModel = None,
    mc_correct: list[int] = None,
    item_categories: list[int] = None
) -> torch.Tensor:
    """
    Impute missing values.

    Parameters
    ----------
    data : torch.Tensor
        A 2D tensor where each row is a response vector and each column is an item.
    method : str, optional
        The imputation method to use. Options are 'zero', 'mean', 'random_incorrect'. (default is 'zero')

        - 'zero': Impute missing values with 0.
        - 'mean': Impute missing values with the item means.
        - 'random incorrect': Impute missing values with a random incorrect response. This method is only valid for multiple choice data.
        - 'prior expected': Impute missing values with the expected scores for the latent space prior distribution mean.
    model : BaseIRTModel, optional
        Only for method='random_incorrect' or 'prior expected'. The IRT model to use for imputation. (default is None)
    mc_correct : list[int], optional
        Only for method='random_incorrect'. A list of integers where each integer is the correct response for the corresponding item. If None, the data is assumed to be non multiple choice (or dichotomously scored multiple choice with only 0's and 1's). (default is None)
    item_categories : list[int], optional
        Only for method='random_incorrect'. A list of integers where each integer is the number of possible responses for the corresponding item. If None, the number of possible responses is calculated from the data. (default is None)
    """
    imputed_data = data.clone()

    if (imputed_data == -1).any():
        imputed_data[imputed_data == -1] = torch.nan

    if method == "zero":
        imputed_data = torch.where(torch.isnan(imputed_data), torch.tensor(0.0), imputed_data)
    elif method == "mean":
        means = imputed_data.nanmean(dim=0)
        mask = torch.isnan(imputed_data)
        imputed_data[mask] = means.repeat(data.shape[0], 1)[mask]
    elif method == "random incorrect":
        if model is not None:
            if model.mc_correct is not None:
                raise ValueError("The model provided must be a multiple choice item model when using random_incorrect imputation")
            item_categories = model.item_categories
            mc_correct = model.mc_correct
        else:
            if mc_correct is None:
                raise ValueError("mc_correct must be provided when using random_incorrect imputation without a model")
            if item_categories is None:
                item_categories = (torch.where(~imputed_data.isnan(), data, torch.tensor(float('-inf'))).max(dim=0).values + 1).int().tolist()

        for col in range(imputed_data.shape[1]):
            # Get the incorrect non-missing responses from the column
            incorrect_responses = torch.arange(0, item_categories[col], device=imputed_data.device).float()
            incorrect_responses = incorrect_responses[incorrect_responses != mc_correct[col]]
            # Find the indices of missing values in the column
            missing_indices = (imputed_data[:, col].isnan()).squeeze()
            # randomly sample from the incorrect responses and replace missing
            imputed_data[missing_indices, col] = incorrect_responses[torch.randint(0, incorrect_responses.size(0), (missing_indices.sum(),))]
    elif method == "prior expected":
        if model is None:
            raise ValueError("The model must be provided when using prior mean imputation")
        if model.mc_correct is not None:
            raise ValueError("The model provided must be a non-multiple choice item model when using prior mean imputation")
        prior_scores = model.expected_scores(
            torch.zeros(1, model.latent_variables).to(next(model.parameters()).device),
            return_item_scores=True
        ).round()
        mask = torch.isnan(imputed_data)
        imputed_data[mask] = prior_scores.repeat(imputed_data.shape[0], 1).to(
            next(model.parameters()).device
        )[mask]
    else:
        raise ValueError(
            f"{method} imputation is not implmented"
        )

    return imputed_data

@torch.no_grad()
def imw(
    model1: BaseIRTModel,
    model2: BaseIRTModel,
    data: torch.Tensor,
    model1_theta: torch.Tensor = None,
    model2_theta: torch.Tensor = None,
    max_iterations = 100,
    **kwargs
):
    r"""
    Quantifies the improvement in model predictions from model 2 over model 1 by computing the InterModel Vigorish (IMW) metric by :cite:t:`Domingue2024`.
    The IMW is the expected gain if one were to place fair bets on model 1 making correct predictions on the observed data,
    but the actual predictions are made by model 2.
    
    Parameters
    ----------
    model1 : BaseIRTModel
        The first model to compare.
    model2 : BaseIRTModel
        The second model to compare.
    data : torch.Tensor, optional
        The data to use for evaluating the models. Typically not the same as the model training data to avoid overfitting.
    model1_theta : torch.Tensor, optional
        The latent scores for the first model. If None, the latent scores are calculated using the first model. (default is None)
    model2_theta : torch.Tensor, optional
        The latent scores for the second model. If None, the latent scores are calculated using the second model. (default is None)
    max_iterations : int, optional
        The maximum number of iterations for the Newton-Raphson method. (default is 100)
    **kwargs
        Additional keyword arguments to pass to the model evaluate method.

    Returns
    -------
    torch.Tensor
        The IMW value. A positive value indicates that model 2 predicts better than model 1.

    Examples
    --------
    >>> import irtorch
    >>> from irtorch.estimation_algorithms import MML
    >>> from irtorch.models import OneParameterLogistic, TwoParameterLogistic, ThreeParameterLogistic
    >>> data = irtorch.load_dataset.swedish_sat_binary()[:, :80]
    >>> train_data, test_data = irtorch.split_data(data, 0.8)
    >>> model1 = OneParameterLogistic(train_data)
    >>> model2 = TwoParameterLogistic(train_data)
    >>> model3 = ThreeParameterLogistic(train_data)
    >>> model1.fit(train_data, MML())
    >>> model2.fit(train_data, MML())
    >>> model3.fit(train_data, MML())
    >>> imw12 = irtorch.utils.imw(model1, model2, test_data)
    >>> imw23 = irtorch.utils.imw(model2, model3, test_data)
    >>> imw13 = irtorch.utils.imw(model1, model3, test_data)
    >>> print(f"IMW 1-2: {imw12:.4f}")
    >>> print(f"IMW 2-3: {imw23:.4f}")
    >>> print(f"IMW 1-3: {imw13:.4f}")
    """
    # Same as the log of the geometric mean of the likelihoods
    log_geometric_mean1 = model1.evaluate.log_likelihood(data, model1_theta, reduction="mean", **kwargs)
    log_geometric_mean2 = model2.evaluate.log_likelihood(data, model2_theta, reduction="mean", **kwargs)
    # Probability of the entropy that leads to the geometric mean
    log_geometric_means = torch.stack((log_geometric_mean1, log_geometric_mean2))
    w = _imw_newton_raphson(log_geometric_means, max_iter=max_iterations)
    imw = (w[1]-w[0])/w[0]
    return imw

def _imw_newton_raphson(negative_entropy, max_iter=100, tol=1e-6):
    negative_entropy = negative_entropy.clone().detach()
    w = torch.full_like(negative_entropy, 0.75)
    for _ in range(max_iter):
        fx = w * torch.log(w) + (1 - w) * torch.log(1 - w) - negative_entropy
        fpx = torch.log(w) - torch.log(1 - w)
        
        # Avoid division by zero
        fpx = torch.where(fpx == 0, torch.tensor(1e-6), fpx)
        
        delta_x = -fx / fpx
        w = w + delta_x
        w = torch.clamp(w, 0.5 + 1e-6, 1 - 1e-6)
        if torch.max(torch.abs(delta_x)) < tol:
            break
    return w


def fit_multiple_models_cpu(
    models: list[BaseIRTModel],
    train_data: torch.Tensor,
    cores_to_use: int = None,
    **kwargs
) -> list[BaseIRTModel]:
    """
    Train multiple models on the same data using multiprocessing.

    Parameters
    ----------
    models : BaseIRTModel
        The IRT models to train. Note that these should be untrained model instances.
    train_data : torch.Tensor
        The data to use for training.
    cores_to_use : int, optional
        Number of cores to use for multiprocessing when device is 'cpu'. (default uses one per model)
    **kwargs
        Additional keyword arguments to pass to the model fit method.

    Returns
    -------
    list[BaseIRTModel]
        A list of trained models.

    Examples
    --------
    >>> from irtorch.models import GeneralizedPartialCredit
    >>> from irtorch.estimation_algorithms import VAE
    >>> from irtorch.load_dataset import swedish_national_mathematics_1
    >>> from irtorch.utils import train_multiple_models_cpu
    >>> data = swedish_national_mathematics_1()
    >>> # train 3 models with 1, 2 and 3 latent variables
    >>> models = [GeneralizedPartialCredit(latent_variables = i+1, data=data) for i in range(3)]
    >>> if __name__ == '__main__':
    ...     train_multiple_models_cpu(models, data, algorithm = VAE())
    """
    if cores_to_use is None:
        cores_to_use = min(len(models), mp.cpu_count())

    print(f"Using {cores_to_use} cores")
    original_threads = torch.get_num_threads()
    try:
        # Ensure each subprocess uses a single thread for PyTorch.
        torch.set_num_threads(1)
        with mp.Pool(cores_to_use) as pool:
            trained_models = pool.starmap(_fit_model, [(model, train_data, kwargs) for model in models])
    finally:
        torch.set_num_threads(original_threads)

    return trained_models

def _fit_model(model: BaseIRTModel, train_data: torch.Tensor, kwargs):
    """
    This is a worker function used in multiprocessing
    """
    model.fit(train_data, device="cpu", **kwargs)
    return model

def split_data(data, train_ratio=0.8, shuffle=True):
    """
    Splits a tensor into training and testing datasets.

    Parameters
    ----------
    data : torch.Tensor
        The dataset to be split. It should be a tensor where the first dimension corresponds to the number of samples.
    train_ratio : float, optional
        The proportion of the dataset to include in the train split. This should be a decimal representing
        the percentage of data used for training (e.g., 0.8 for 80% training and 20% testing). The default is 0.8.
    shuffle : bool, optional
        Whether to shuffle the data before splitting. It is highly recommended to shuffle the data
        to ensure that the training and testing datasets are representative of the overall dataset. The default is True.

    Returns
    -------
    tuple[torch.Tensor, torch.Tensor]
        A tuple containing two tensors. The first tensor is the training dataset, and the second tensor is the testing dataset.
        The split is based on the `train_ratio` parameter, and if `shuffle` is True, the split is performed on the shuffled data.

    Examples
    --------
    >>> data = torch.rand(100, 10)  # Example tensor with 100 samples and 10 features each
    >>> train_data, test_data = split_tensor(data, train_ratio=0.8, shuffle=True)
    >>> print("Training data size:", train_data.shape)
    >>> print("Testing data size:", test_data.shape)
    """
    if shuffle:
        indices = torch.randperm(data.size(0))
        data_shuffled = data[indices]
    else:
        data_shuffled = data

    train_size = int(data.size(0) * train_ratio)
    train_data = data_shuffled[:train_size]
    test_data = data_shuffled[train_size:]

    return train_data, test_data

def set_seed(seed: int):
    """
    Seed all random number generators for reproducibility (random, numpy and torch).

    Parameters
    ----------
    seed : int
        The seed value to use for seeding the random number generators.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # for multi-GPU, but also works for single.
    # CuDNN (CUDA Deep Neural Network library) provides highly optimized implementations for many neural network operations.
    # Some of these implementations are non-deterministic due to parallelism and optimization techniques.
    # Setting torch.backends.cudnn.deterministic = True ensures that CuDNN selects deterministic algorithms, eliminating randomness from these operations.
    torch.backends.cudnn.deterministic = True
    # CuDNN's benchmarking feature selects the most efficient algorithm for your specific hardware configuration, but introduces randomness
    torch.backends.cudnn.benchmark = False



import logging
import torch
from torch.distributions import Normal
from irtorch.models import BaseIRTModel
from irtorch.estimation_algorithms.ae import AE
from irtorch.estimation_algorithms.encoders import VariationalEncoder
from irtorch.irt_dataset import PytorchIRTDataset

logger = logging.getLogger("irtorch")

class VAE(AE):
    """
    Variational autoencoder neural network with importance weighted sampling for fitting IRT models :cite:p:`Urban2021`.
    This method is effective when fitting high-dimensional IRT models with large datasets.

    """
    def __init__(self):
        super().__init__()
        self.iw_samples = 5
        self.annealing_iterations = 5
        self.anneal = True
        self.annealing_factor = 1.0

    def fit(
        self,
        model: BaseIRTModel,
        train_data: torch.Tensor,
        one_hot_encoded: bool = True,
        imputation_method: str = None,
        learning_rate: float = 0.002,
        learning_rate_updates_before_stopping: int = 2,
        evaluation_interval_size: int = 60,
        max_epochs: int = 10000,
        batch_size: int = None,
        batch_normalization_encoder: bool = False,
        nonlinear_encoder = torch.nn.ELU(),
        hidden_layers_encoder: list[int] = None,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        anneal: int = True,
        annealing_iterations: int = 5,
        iw_samples: int = 5,
    ):
        """
        Train an IRT model using the variational autoencoder. If the algorithm fails to converge, try lowering the learning rate. Use batch_size if the data is too large to fit in memory.

        Parameters
        ----------
        model : BaseIRTModel
            The model to fit. Needs to inherit :class:`irtorch.models.BaseIRTModel`.
        train_data : torch.Tensor
            The training data. Item responses should be coded 0, 1, ... and missing responses coded as nan or -1.
        one_hot_encoded : bool, optional
            Whether the model uses one-hot encoded data. (default is False)
        imputation_method : str, optional
            The method to use for imputing missing data for the encoder. For options see :func:`irtorch.utils.impute_missing`.
            Only methods not relying on a fitted model can be used. 
            Note that missing values are removed from the loss calculation even after imputation.
            If you do not want this, do the imputation to your dataset before fitting. (default is None and only works for one-hot encoded inputs)
        learning_rate : float, optional
            The initial learning rate for the optimizer. (default is 0.002)
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training. (default is 2)
        evaluation_interval_size: int, optional
            The number of iterations between each model evaluation during training. (default is 80)
        max_epochs : int, optional
            The maximum number of epochs to train for. (default is 1000)
        batch_size : int, optional
            The batch size for training. (default is 64)
        batch_normalization_encoder : bool, optional
            Whether to use batch normalization for the encoder. (default is True)
        nonlinear_encoder : torch.nn.Module, optional
            The non-linear function to use after each hidden layer in the encoder. (default is torch.nn.ELU())
        hidden_layers_encoder : list[int], optional
            List of hidden layers for the encoder. Each element is a layer with the number of neurons represented as integers. If not provided, uses one hidden layer with 2 * sum(item_categories) neurons.
        device : str, optional
            The device to run the model on. (default is "cuda" if available else "cpu".)
        anneal : bool, optional
            Whether to anneal the KL divergence. (default is True)
        annealing_iterations : int, optional
            The number of iterations to anneal the KL divergence. (default is 5)
        iw_samples : int, optional
            The number of importance weighted samples to use. (default is 5)
        """
        if self.train_data is not None:
            self.train_data = torch.cat((self.train_data, train_data), dim=0).contiguous()
        else:
            self.train_data = train_data.contiguous()

        # Re-initialize the training history
        self.training_history = {
            "train_loss": [],
        }
        self.best_model_state = None
        self.batch_mean_losses = []
        self.best_avg_loss = float("Inf"), 0
        self.total_iterations = 0
        self.lr_update_count = 0
        self.evaluation_interval_size = evaluation_interval_size

        self.iw_samples = iw_samples
        self.annealing_iterations = annealing_iterations
        self.anneal = anneal
        self.one_hot_encoded = one_hot_encoded

        if batch_size is None:
            batch_size = train_data.shape[0]
        if one_hot_encoded:
            input_dim = sum(model.item_categories)
        else:
            input_dim = len(model.item_categories)
        if hidden_layers_encoder is None:  # 1 layer with 2x number of categories as neurons is default
            hidden_layers_encoder = [2 * sum(model.item_categories)]

        self.encoder = VariationalEncoder(
            input_dim,
            model.latent_variables,
            hidden_layers_encoder,
            batch_normalization=batch_normalization_encoder,
            nonlinear=nonlinear_encoder,
        )

        if not one_hot_encoded and imputation_method is None:
            raise ValueError("imputation_method must be supplied for non-one-hot encoded data.")

        train_data_irt = PytorchIRTDataset(
            data=train_data.to(device),
            one_hot_encoded=self.one_hot_encoded,
            item_categories=model.item_categories,
            imputation_method=imputation_method,
            mc_correct=model.mc_correct
        )
        self.data_loader = torch.utils.data.DataLoader(
            train_data_irt,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=False,
        )

        self.optimizer = torch.optim.Adam(
            list(self.encoder.parameters()) + list(model.parameters()), lr=learning_rate, amsgrad=True
        )

        # Reduce learning rate when loss stops decreasing ("min")
        # we multiply the learning rate by the factor
        # patience: We need no improvement after x epochs for it to trigger
        self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, mode="min", factor=0.6, patience=1
        )

        self.encoder.to(device)
        model.to(device)
        self._training_loop(model, max_epochs, learning_rate_updates_before_stopping)
        self.encoder.to("cpu")
        model.to("cpu")
        self.encoder.eval()
        model.eval()

        # store the latent theta scores of the training data
        # used for more efficient computation when using other methods
        self.training_theta_scores = self.theta_scores(
            train_data_irt.input_data.to("cpu")
        ).clone().detach()

    def _train_batch(self, model: BaseIRTModel, input_batch: torch.Tensor, batch: torch.Tensor, missing_mask: torch.Tensor):
        """
        Train the model on a batch of data.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        input_batch : torch.Tensor
            The input batch for the encoder.
        batch : torch.Tensor
            The batch of data for computing the loss.
        missing_mask : torch.Tensor
            The mask for missing data.

        Returns
        -------
        tuple
            The loss after training on the batch.
        """
        mean, logvar = self.encoder(input_batch)
        if torch.isnan(mean).all():
            return torch.tensor(torch.nan)

        # takes iw_samples from the latent space for each data point (for importance weighting)
        mean = mean.repeat(self.iw_samples, 1)
        logvar = logvar.repeat(self.iw_samples, 1)
        theta_samples = self.reparameterize(mean, logvar)

        batch_logits = model(theta_samples)

        batch_loss = self._loss_function(model, batch, missing_mask, batch_logits, theta_samples, mean, logvar)
        return batch_loss

    def reparameterize(self, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mean + eps * std

    # IWAE Loss = E log (1/K Σ_{k=1}^{K} w_k) (ELBO for 1 iw sample)
    def _loss_function(
        self,
        model: BaseIRTModel,
        data: torch.Tensor,
        missing_mask: torch.Tensor,
        logits: torch.Tensor,
        theta_samples: torch.Tensor,
        mean: torch.Tensor,
        logvar: torch.Tensor,
    ):
        """
        The IWAE loss function for the model as introduced by Burda et. al. (2015).
        When self.iw_samples == 1 this generalizes to the standard VAE ELBO loss.

        Parameters
        ----------
        model : BaseIRTModel
            The model to fit.
        data : torch.Tensor
            The input data.
        missing_mask : torch.Tensor
            The mask for missing data.
        logits : torch.Tensor
            The logits output by the model.
        theta_samples : torch.Tensor
            Samples from the encoder distribution
        mean : torch.Tensor
            Means from the encoder
        logvar : torch.Tensor
            Log of the variance from the encoder

        Returns
        -------
        torch.Tensor
            The calculated loss.
        """
        if self.anneal:
            self.annealing_factor = min(1.0, self.total_iterations / self.annealing_iterations)
        else:
            self.annealing_factor = 1.0

        log_p_x_theta = model.log_likelihood(
            data.repeat(self.iw_samples, 1),
            logits,
            missing_mask.repeat(self.iw_samples, 1),
            loss_reduction="none",
        )

        # Reshape tensor by each iw-sample
        # we need to keep a dimension for each respondent, but we can sum over items
        log_p_x_theta = log_p_x_theta.view(
            self.iw_samples,
            log_p_x_theta.shape[0] // (self.iw_samples * data.shape[1]),
            data.shape[1],
        ).nansum(2)

        if self.iw_samples == 1:
            # ELBO (1 sample) can be computed using the true means and variances
            # sum up kl div for each person (row)
            kl_div = 0.5 * torch.sum(mean.pow(2) + logvar.exp() - 1 - logvar, dim=1)
            iwae_bound = log_p_x_theta - self.annealing_factor * kl_div
        else:
            log_p_theta = Normal(0, 1).log_prob(theta_samples)
            log_q_theta_x = Normal(mean, torch.exp(0.5 * logvar)).log_prob(theta_samples)
            kl_div = log_q_theta_x - log_p_theta
            kl_div = kl_div.view(
                self.iw_samples, kl_div.shape[0] // self.iw_samples, kl_div.shape[1]
            ).sum(2)

            iwae_bound = log_p_x_theta - self.annealing_factor * kl_div
            if self.iw_samples > 1:
                iwae_bound = torch.logsumexp(iwae_bound, dim=0) - torch.log(
                    torch.tensor(self.iw_samples).float()
                )  # Importance weighting log (1/K Σ_{k=1}^{K} w_k)

        # Estimate expectation
        return -iwae_bound.mean()

    def _batch_fit_measures(self, model: BaseIRTModel, input_batch: torch.Tensor, batch: torch.Tensor, missing_mask: torch.Tensor):
        """
        Calculate the fit measures for a batch.

        Parameters
        ----------
        model : BaseIRTModel
            The model to fit.
        batch : torch.Tensor
            The batch of data.

        Returns
        -------
        tuple
            The loss, log likelihood, and accuracy for the batch.
        """
        encoder_mean, _ = self.encoder(input_batch)
        output = model(encoder_mean)
        log_likelihood = model.log_likelihood(
            batch, output, missing_mask
        )
        return -log_likelihood / batch.shape[0]

    @torch.no_grad()
    def theta_scores(
        self,
        data: torch.Tensor,
    ):
        """
        Get the latent scores from an input dataset using the encoder.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents.
        
        Returns
        -------
        torch.Tensor
            A 2D tensor of latent scores. Rows are respondents and columns are latent variables.
        """
        data = data.contiguous()
        return self.encoder(data)[0]

    @torch.no_grad()
    def sample_latent_variables(
        self, model: BaseIRTModel, sample_size: int, input_data: torch.Tensor = None
    ):
        """
        Sample latent variables from the encoder.

        Parameters
        ----------
        model : BaseIRTModel
            The model.
        sample_size : int
            The number of samples to draw.
        input_data : torch.Tensor, optional
            The data to sample from. If None, uses the training data. (default is None)

        Returns
        -------
        torch.Tensor
            A 2D tensor of latent variables. Rows are samples and columns are latent variables.
        """
        if input_data is None:
            input_data = self.train_data
        else:
            input_data = input_data.contiguous().to(next(model.parameters()).device)

        input_data_irt = PytorchIRTDataset(
            data=input_data,
            one_hot_encoded=self.one_hot_encoded,
            item_categories=model.item_categories,
            imputation_method=self.imputation_method,
            mc_correct=model.mc_correct
        )

        # Sample test scores until we have sample_size
        indices = torch.randint(low=0, high=input_data.size(0), size=(sample_size,)).to(
            next(model.parameters()).device
        )
        samples = torch.index_select(input_data_irt.input_data, 0, indices)
        # run the sample through the encoder
        mean, logvar = self.encoder(samples)
        return self.reparameterize(mean, logvar)

    @torch.no_grad()
    def latent_credible_interval(
        self, input_data: torch.Tensor, alpha=0.05
    ) -> torch.tensor:
        input_data = input_data.contiguous()
        mean, logvar = self.encoder(input_data)
        # Create a Gaussian distribution with the calculated mean and std
        dist = Normal(mean, torch.exp(0.5 * logvar))

        # Prepare alpha tensors of the same shape as mean and std
        lower_alpha = torch.full_like(mean, alpha / 2)
        upper_alpha = torch.full_like(mean, 1 - alpha / 2)

        # Calculate the lower and upper bounds of the credible interval
        lower = dist.icdf(
            lower_alpha.clone().detach()
        )
        upper = dist.icdf(
            upper_alpha.clone().detach()
        )

        return lower, mean, upper

    @torch.no_grad()
    def latent_mean_se(self, input_data: torch.Tensor) -> torch.tensor:
        """
        Get the posterior mean and standard error of the latent variables.

        Parameters
        ----------
        input_data : torch.Tensor
            The input data.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor]
            The posterior mean and standard error of the latent variables.
        """
        input_data = input_data.contiguous()
        mean, logvar = self.encoder(input_data)
        return mean, torch.exp(0.5 * logvar)



import torch
import torch.nn as nn
from torch.distributions import Distribution
from irtorch.torch_modules import RationalQuadraticSpline

class NeuralSplineFlow(nn.Module):
    """
    Normalizing flow using rational-quadratic splines :cite:p:`Durkan2019`.

    Parameters
    ----------
    transform: RationalQuadraticSpline
        Transforms values from the observed data to the chosen distribution.
    distribution: Distribution 
        The base distribution of the flow that generates the observed data.
    """
    def __init__(self, transformation: RationalQuadraticSpline, distribution: Distribution):
        super().__init__()
        self._transformation = transformation
        self._distribution = distribution

    def forward(self, inputs) -> torch.Tensor:
        """
        Transforms from observed data into the chosen distribution (noise).

        Parameters
        ----------
        inputs: torch.Tensor
            A tensor of shape [batch_size, ...], the data to be transformed.

        Returns
        -------
        torch.Tensor
            A tensor of shape [batch_size, ...].
        """
        noise, _ = self._transformation(inputs, inverse=False)
        return noise

    def log_prob(self, value) -> torch.Tensor:
        """
        The log-likelihood of the observed data in the chosen distribution after transformation. Usually the negative loss when training.

        Parameters
        ----------
        value: torch.Tensor
            A tensor of shape [batch_size, ...], the data to be transformed.

        Returns
        -------
        torch.Tensor
            A tensor of shape [batch_size, ...].
        """
        noise, logabsdet = self._transformation(value, inverse=False)
        log_prob = self._distribution.log_prob(noise)
        return log_prob + logabsdet

    def _sample_obs(self, num_samples) -> torch.Tensor:
        noise = self._distribution.sample(num_samples)
        samples, _ = self._transformation(noise, inverse=True)
        return samples

    def inverse(self, inputs) -> torch.Tensor:
        """
        Transforms from the chosen distribution (noise) into observed data.

        Parameters
        ----------
        inputs: torch.Tensor
            A tensor of shape [batch_size, ...], the data to be transformed.

        Returns
        -------
        torch.Tensor
            A tensor of shape [batch_size, ...].
        """
        noise, _ = self._transformation(inputs, inverse=True)
        return noise



import pandas as pd
import torch
import torch.nn as nn
from irtorch.models.base_irt_model import BaseIRTModel

class OneParameterLogistic(BaseIRTModel):
    r"""
    One parametric logistic (1PL) IRT model, also known as the Rasch model :cite:p:`Rasch1960`. This model is only available for unidimensional latent variables.

    Parameters
    ----------
    data: torch.Tensor, optional
        A 2D torch tensor with test data. Used to automatically compute the number of items. Columns are items and rows are respondents. (default is None)
    items : int
        Number of items.
    
    Notes
    -----
    For an item :math:`j`, the model defines the probability for responding correctly as:

    .. math::

        \frac{
            \exp(\theta + d_j)
        }{
            1+\exp(\theta + d_j)
        },

    where:

    - :math:`\theta` is the latent variable.
    - :math:`d_j` is the bias term for item :math:`j`.

    Examples
    --------
    >>> from irtorch.models import OneParameterLogistic
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.load_dataset import swedish_sat_binary
    >>> # Use quantitative part of the SAT data
    >>> data = swedish_sat_binary()[:, :80]
    >>> model = OneParameterLogistic(items=80)
    >>> model.fit(train_data=data, algorithm=AE())
    """
    def __init__(
        self,
        data: torch.Tensor = None,
        items: int = None,
    ):
        if items is None and data is None:
            raise ValueError("Either items or data must be provided to initialize the model.")
        if data is not None:
            items = data.size(1)
        super().__init__(latent_variables=1, item_categories = [2] * items)

        self.output_size = self.items * 2
        self.bias_param = nn.Parameter(torch.zeros(self.items))

        first_category = torch.zeros(self.items, 2)
        first_category[:, 0] = 1.0
        first_category = first_category.reshape(-1)

        free_bias = 1 - first_category
        self.register_buffer("free_bias", free_bias.bool())
        self.register_buffer("first_category", first_category.bool())
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.zeros_(self.bias_param)
    
    def forward(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model.

        Parameters
        ----------
        theta : torch.Tensor
            2D tensor with latent variables. Rows are respondents and latent variables are columns. 

        Returns
        -------
        output : torch.Tensor
            2D tensor. Rows are respondents and columns are item category logits.
        """
        bias = torch.zeros(self.output_size, device=theta.device)
        bias[self.free_bias] = self.bias_param
        output = theta + bias
        output[:, self.first_category] = 0

        return output

    def item_parameters(self) -> pd.DataFrame:
        """
        Get the item parameters for a fitted model.

        Returns
        -------
        pd.DataFrame
            A dataframe with the item parameters.
        """
        biases = self.bias_param
        biases_df = pd.DataFrame(biases.reshape(-1, 1).detach().numpy())
        biases_df.columns = ["d"]

        return biases_df

    @torch.no_grad()
    def item_theta_relationship_directions(self, theta:torch.Tensor = None) -> torch.Tensor:
        """
        Get the relationship directions between each item and latent variable for a fitted model.

        Parameters
        ----------
        theta : torch.Tensor, optional
            Not needed for this model. (default is None)
            
        Returns
        -------
        torch.Tensor
            A 2D tensor with the relationships between the items and latent variables. Items are rows and latent variables are columns.
        """
        return torch.ones(self.items, 1)
    


from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import torch
import pandas as pd
import numpy as np
from torch.distributions import MultivariateNormal
from irtorch.estimation_algorithms import VAE, AE, MML
from irtorch.quantile_mv_normal import QuantileMVNormal
from irtorch.gaussian_mixture_model import GaussianMixtureModel
from irtorch._internal_utils import (
    conditional_score_distribution,
    sum_incorrect_probabilities,
    get_missing_mask,
    correlation_matrix,
    dynamic_print,
    entropy,
    joint_entropy_matrix
)

if TYPE_CHECKING:
    from irtorch.models.base_irt_model import BaseIRTModel

logger = logging.getLogger("irtorch")

class Evaluator:
    """
    Class for evaluating IRT model performance using various metrics.
    A fitted :doc:`model <irt_models>` typically holds an instance of this class in its `evaluation` property. 
    Thus the methods can be accessed through `model.evaluation.method_name()`.

    Parameters
    ----------
    model : BaseIRTModel
        The IRT model to evaluate.
    """
    def __init__(self, model: BaseIRTModel):
        self.model = model
        self.latent_density = None

    def _evaluate_data_theta_input(
            self,
            data: torch.Tensor = None,
            theta: torch.Tensor = None,
            **kwargs
        ):
        """"
        Helper function for evaluating the data and theta inputs for various performance measure methods.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data. If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.
        """
        if data is None:
            data = self.model.algorithm.train_data
        else:
            data = data.contiguous()

        if theta is None:
            theta = self.model.latent_scores(
                data=data, rescale=False, **kwargs)

        missing_mask = get_missing_mask(data)

        return data, theta, missing_mask


    @torch.no_grad()
    def approximate_latent_density(
        self,
        theta_scores: torch.Tensor,
        approximation: str = "qmvn",
        cv_n_components: list[int] = None,
    ) -> None:
        """
        Approximate the latent space density.

        Parameters
        ----------
        theta_scores : torch.Tensor
            A 2D tensor with theta scores. Each row represents one respondent, and each column an item.
        approximation : str, optional
            The approximation method to use. (default is 'qmvn')
            - 'qmvn' for quantile multivariate normal approximation of a multivariate joint density function (QuantileMVNormal class).
            - 'gmm' for a gaussian mixture model.

        cv_n_components: list[int], optional
            The number of components to use for cross-validation with Gaussian Mixture Models. (default is [2, 3, 4, 5, 10])

        Returns
        -------
        torch.Tensor
            A 2D tensor of latent scores, with latent variables as columns.
        """
        if approximation == "gmm":
            cv_n_components = [2, 3, 4, 5, 10] if cv_n_components is None else cv_n_components
            self.latent_density = self._cv_gaussian_mixture_model(theta_scores, cv_n_components)
        elif approximation == "qmvn":
            self.latent_density = QuantileMVNormal()
            self.latent_density.fit(theta_scores)
        else:
            raise ValueError("Invalid approximation method. Choose either 'qmvn' or 'gmm'.")

    def _cv_gaussian_mixture_model(self, data: torch.Tensor, cv_n_components: list[int]) -> GaussianMixtureModel:
        if len(cv_n_components) > 1:
            logger.info("Performing cross-validation for Gaussian Mixture Model components...")
            n_folds = 5
            average_log_likelihood = torch.zeros(len(cv_n_components))
            for comp_ind, components in enumerate(cv_n_components):
                log_likelihoods = torch.zeros(n_folds)
                data_chunks = data.chunk(n_folds)
                for i, data_val in enumerate(data_chunks):
                    train_chunks = [x for j, x in enumerate(data_chunks) if j != i]
                    data_train = torch.cat(train_chunks, dim=0)

                    gmm = GaussianMixtureModel(n_components=components, n_features=data.shape[1])
                    gmm.fit(data_train)

                    # Compute the log likelihood on the validation data
                    log_likelihoods[i] = gmm(data_val)

                # Compute the average log likelihood over the folds
                average_log_likelihood[comp_ind] = log_likelihoods.mean()

            # Select the number of components that maximizes the average log likelihood
            optimal_n_components = cv_n_components[average_log_likelihood.argmax()]
            logger.info("Fitting Gaussian Mixture Model for latent variable density.")
            gmm = GaussianMixtureModel(n_components=optimal_n_components, n_features=data.shape[1])
        else:
            logger.info("Fitting Gaussian Mixture Model for latent variable density.")
            gmm = GaussianMixtureModel(n_components=cv_n_components[0], n_features=data.shape[1])

        gmm.fit(data)
        return gmm

    @torch.no_grad()
    def residuals(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        average_over: str = "none",
        **kwargs
    ) -> torch.Tensor:
        """
        Compute model residuals using the supplied data. 
        
        For multiple choice models, the residuals are computed as 1 - the probability of the selected response option.
        For other models, the residuals are computed as the difference between the observed and model expected item scores.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        average_over: str = "none", optional
            Whether to average the residuals and over which level. Can be 'everything', 'items', 'respondents' or 'none'. Use 'none' for no average. For example, with 'respondent' the residuals are averaged over all respondents and is thus an average per item. (default is 'none')
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.
            
        Returns
        -------
        torch.Tensor
            The residuals.
        """
        data, theta, _ = self._evaluate_data_theta_input(data, theta, **kwargs)

        missing_mask = get_missing_mask(data)
        data[torch.isnan(data)] = -1
        if self.model.mc_correct is not None:
            # 3D tensor with dimensions (respondents, items, item categories)
            probabilities = self.model.item_probabilities(theta)
            # Creating a range tensor for slice indices
            respondents = torch.arange(probabilities.size(0)).view(-1, 1)
            # Expand slices to match the shape of indices
            expanded_respondents = respondents.expand_as(data)
            model_probs = probabilities[expanded_respondents, torch.arange(probabilities.size(1)), data.int()]
            residuals = 1 - model_probs
        else:
            residuals = data - self.model.expected_scores(theta, return_item_scores=True)

        residuals[missing_mask] = float("nan")
        if average_over == "items":
            return residuals.nanmean(dim=1)
        if average_over == "respondents":
            return residuals.nanmean(dim=0)
        if average_over == "everything":
            return residuals.nanmean(dim=None)

        return residuals

    @torch.no_grad()
    def group_fit_residuals(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        latent_variable: int = 1,
        standardize: bool = True,
        groups: int = 10,
        theta_estimation: str = "ML",
        rescale: bool = True,
        **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Group the respondents based on their ordered latent variable scores.
        Calculate the residuals between the model estimated and observed data within each group.
        See :cite:t:`VanderLinden2016`, Chapter 20 for more details.
        

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        theta : torch.Tensor, optional
            A 2D tensor containing the pre-estimated theta scores for each respondent in the data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        latent_variable: int, optional
            Specifies the latent variable along which ordering and grouping should be performed. (default is 1)
        standardize : bool, optional
            Specifies whether the residuals should be standardized. (default is True)
        groups: int
            The number of groups. (default is 10)
        theta_estimation : str, optional
            Method used to obtain the theta scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'NN')
        rescale : bool, optional
            Whether to compute the latent scores on the theta transformation scale if it exists. (default is True)
        **kwargs : dict, optional
            Additional keyword arguments used for scale computation. Refer to documentation for the chosen scale in the :doc:`scales` documentation section for additional details.
            
        Returns
        -------
        tuple[torch.Tensor, torch.Tensor]
            A tuple with torch tensors. The first one holds the residuals for each group and has dimensions (groups, items, item categories). The second one is a 1D tensor and holds the mid points of the groups.
        """
        grouped_data_probabilties, grouped_model_probabilties, group_mid_points = \
            self.latent_group_probabilities(
                data=data,
                theta=theta,
                rescale=rescale,
                latent_variable=latent_variable,
                groups=groups,
                theta_estimation=theta_estimation,
                **kwargs
            )

        raw_residuals = grouped_data_probabilties - grouped_model_probabilties
        if standardize:
            # get number of people in each group
            data = torch.chunk(data, groups)
            group_n = torch.tensor([group.shape[0] for group in data])

            std_residuals = raw_residuals / (grouped_model_probabilties * (1 - grouped_model_probabilties) / group_n.reshape(-1, 1, 1)).sqrt()
            return std_residuals, group_mid_points
        
        for item, categories in enumerate(self.model.item_categories):
            # set non-existing categories to nan
            raw_residuals[:, item, categories:] = float("nan")

        return raw_residuals, group_mid_points
    
    @torch.no_grad()
    def accuracy(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        level: str = "all",
        **kwargs
    ):
        """
        Calculate the prediction accuracy of the model for the supplied data.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        level: str = "all", optional
            Specifies the level at which the accuracy is calculated. Can be 'all', 'item' or 'respondent'. For example, for 'item' the accuracy is calculated for each item. (default is 'all')
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.

        Returns
        -------
        torch.Tensor
            The accuracy.
        """
        data, theta, missing_mask = self._evaluate_data_theta_input(data, theta, **kwargs)

        probabilities = self.model.item_probabilities(theta)
        accuracy = (torch.argmax(probabilities, dim=2) == data).float()
        accuracy[missing_mask] = float("nan")

        if level == "item":
            dim = 0
        elif level == "respondent":
            dim = 1
        else:
            dim = None
        
        return accuracy.nanmean(dim=dim)

    @torch.no_grad()
    def infit_outfit(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        level: str = "item",
        **kwargs
    ):
        """
        Calculate person or item infit and outfit statistics. These statistics help identifying items that do not behave as expected according to the model
        or respondents with unusual response patterns. Items that do not behave as expectedly can be reviewed for possible revision or removal 
        to improve the overall test quality and reliability. Respondents with unusual response patterns can be reviewed for possible cheating or other issues.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        level: str = "item", optional
            Specifies whether to compute item or respondent statistics. Can be 'item' or 'respondent'. (default is 'item')
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.

        Returns
        -------
        torch.Tensor
            The infit statistics.

        Notes
        -----
        Infit and outift are computed as follows:

        .. math::
            \\begin{align}
            \\text{Item j infit} = \\frac{\\sum_{i=1}^{n} (O_{ij} - E_{ij})^2}{\\sum_{i=1}^{n} W_{ij}} \\\\
            \\text{Respondent i infit} = \\frac{\\sum_{j=1}^{J} (O_{ij} - E_{ij})^2}{\\sum_{j=1}^{J} W_{ij}} \\\\
            \\text{Item j outfit} = \\frac{\\sum_{i=1}^{n} (O_{ij} - E_{ij})^2/W_{ij}}{n} \\\\
            \\text{Respondent i outfit} = \\frac{\\sum_{j=1}^{J} (O_{ij} - E_{ij})^2/W_{ij}}{J}
            \\end{align}

        Where:

        - :math:`J` is the number of items,
        - :math:`n` is the number of respondents,
        - :math:`O_{ij}` is the observed score on the :math:`j`-th item from the :math:`i`-th respondent.
        - :math:`E_{ij}` is the expected score on the :math:`j`-th item from the :math:`i`-th respondent, calculated from the IRT model.
        - :math:`W_{ij}` is the weight on the :math:`j`-th item from the :math:`j`-th respondent. This is the variance of the item score :math:`W_{ij}=\\sum^{M_j}_{m=0}(m-E_{ij})^2P_{ijk}` where :math:`M_j` is the maximum item score and :math:`P_{ijk}` is the model probability of a score :math:`k` on the :math:`j`-th item from the :math:`i`-th respondent.
        
        """
        if level not in ["item", "respondent"]:
            raise ValueError("Invalid level. Choose either 'item' or 'respondent'.")

        data, theta, missing_mask = self._evaluate_data_theta_input(data, theta, **kwargs)

        expected_scores = self.model.expected_scores(theta, return_item_scores=True)
        probabilities = self.model.item_probabilities(theta)
        observed_scores = data
        if self.model.mc_correct is not None:
            score_indices = torch.zeros(probabilities.shape[1], probabilities.shape[2])
            score_indices.scatter_(1, torch.tensor(self.model.mc_correct).unsqueeze(1), 1)
            score_indices = score_indices.unsqueeze(0).expand(probabilities.shape[0], -1, -1)
            correct_probabilities = (probabilities*score_indices.int()).sum(dim=2)
            variance = correct_probabilities * (1-correct_probabilities)
            possible_scores = torch.zeros_like(probabilities)
            possible_scores[:, :, 1] = 1
            observed_scores = (data == torch.tensor(self.model.mc_correct)).int()
        else:
            possible_scores = torch.arange(0, probabilities.shape[2]).unsqueeze(0).expand(probabilities.shape[0], probabilities.shape[1], -1)
            variance = ((possible_scores-expected_scores.unsqueeze(2)) ** 2 * probabilities).sum(dim=2)

        mse = (observed_scores - expected_scores) ** 2
        mse[missing_mask] = torch.nan
        variance[missing_mask] = torch.nan
        wmse = mse / variance
        wmse[mse == 0] = 0 # if error is 0, set to 0 in case of 0 variance to avoid nans

        if level == "item":
            infit = mse.nansum(dim=0) / variance.nansum(dim=0)
            outfit = wmse.nanmean(dim=0)
        else:
            infit = mse.nansum(dim=1) / variance.nansum(dim=1)
            outfit = wmse.nanmean(dim=1)
        
        return infit, outfit

    @torch.no_grad()
    def log_likelihood(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        reduction: str = "sum",
        level: str = "all",
        **kwargs
    ):
        """
        Calculate the log-likelihood for the provided data.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None and uses the model's training data)
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        reduction : str, optional
            Specifies the reduction method for the log-likelihood. Can be 'sum', 'none' or 'mean'. (default is 'sum')
        level : str, optional
            For reductions other than 'none', specifies the level at which the log-likelihood is summed/averaged. Can be 'all', 'item' or 'respondent'. For example, for 'item' the log-likelihood is summed/averaged for each item over the respondents. (default is 'all')
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.
            
        Returns
        -------
        torch.Tensor
            The log-likelihood for the provided data.
        """
        data, theta, missing_mask = self._evaluate_data_theta_input(data, theta, **kwargs)

        if reduction != "none":
            if level == "item":
                dim = 0
            elif level == "respondent":
                dim = 1
            else:
                dim = None

        likelihoods = self.model.log_likelihood(
            data,
            self.model(theta),
            missing_mask,
            loss_reduction="none"
        )

        # On unseen data, we may end up with 0 probabilities which will result in -inf likelihoods
        likelihoods[likelihoods==-torch.inf] = -30.0
        likelihoods = likelihoods.view(theta.shape[0], -1)
        if reduction in "mean":
            return likelihoods.nanmean(dim=dim)
        if reduction == "sum":
            return likelihoods.nansum(dim=dim)
        
        return likelihoods

    @torch.no_grad()
    def group_fit_log_likelihood(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        groups: int = 10,
        latent_variable: int = 1,
        **kwargs
    ):
        """
        Group the respondents based on their ordered latent variable scores.
        Calculate the average log-likelihood of the data within each group.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        groups: int
            The number of groups. (default is 10)
        latent_variable: int, optional
            Specifies the latent variable along which ordering and grouping should be performed. (default is 1)
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.

        Returns
        -------
        torch.Tensor
            The average log-likelihood for each group.
        """
        data, theta, missing_mask = self._evaluate_data_theta_input(data, theta, **kwargs)

        indicies = torch.sort(theta[:, latent_variable - 1], dim=0)[1]
        theta = theta[indicies]
        data = data[indicies]
        likelihoods = self.model.log_likelihood(
            data,
            self.model(theta),
            missing_mask,
            loss_reduction="none"
        )
        respondent_likelihoods = likelihoods.view(theta.shape[0], -1).sum(dim=1)
        group_mean_likelihoods = torch.zeros(groups)
        start_index = 0
        for group in range(groups):
            group_size = respondent_likelihoods.shape[0] // groups
            remainder = len(respondent_likelihoods) % groups
            if group < remainder:
                group_size += 1
            group_mean_likelihoods[group] = respondent_likelihoods[start_index: start_index + group_size].mean()
            start_index += group_size
            
        return group_mean_likelihoods
    
    def latent_group_probabilities(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        latent_variable: int = 1,
        groups: int = 10,
        theta_estimation: str = "ML",
        rescale: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Group the respondents based on their ordered latent variable scores.
        Calculate both the observed and IRT model probabilities for each possible item response, within each group.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
        theta : torch.Tensor, optional
            The latent variable theta scores for the provided data. If not provided, they will be computed using theta_estimation. (default is None)
        latent_variable: int, optional
            Specifies the latent variable along which ordering and grouping should be performed. (default is 1)
        groups: int
            The number of groups. (default is 10)
        theta_estimation : str, optional
            Method used to obtain the theta scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'NN')
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        rescale : bool, optional
            Whether to group the latent scores on the theta transformation scale if it exists. Note: for uni-dimensional models, all monotone scale transformations are equivalent in this case. (default is True)
        
        Returns
        -------
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]
            A 3D torch tensor with data group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.

            A 3D torch tensor with model group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.

            The third tensor contains the average latent variable values within each group along the specified latent_variable.
        """
        data, theta, _ = self._evaluate_data_theta_input(data=data, theta=theta, theta_estimation=theta_estimation)

        if rescale and self.model.scale:
            transformed_scores = self.model.transform_theta(theta)
            # Sort based on correct column and get the sorted indices
            _, indices = torch.sort(
                transformed_scores[:, latent_variable - 1], dim=0
            )
            # Use the indices to sort
            transformed_scores = transformed_scores[indices]
            theta = theta[indices]

            grouped_transformed = torch.chunk(transformed_scores, groups)
            grouped_theta = torch.chunk(theta, groups)

            group_mid_points = torch.tensor(
                [
                    group[:, latent_variable - 1].mean()
                    for group in grouped_transformed
                ]
            )
        else:
            _, indices = torch.sort(theta[:, latent_variable - 1], dim=0)
            theta = theta[indices]
            grouped_theta = torch.chunk(theta, groups)
            group_mid_points = torch.tensor(
                [group[:, latent_variable - 1].median() for group in grouped_theta]
            )

        data = data[indices]
        grouped_data = torch.chunk(data, groups)
        grouped_data_probabilties = self._grouped_data_probabilities(grouped_data)
        grouped_model_probabilties = self._grouped_theta_probabilities(grouped_theta)
        return grouped_data_probabilties, grouped_model_probabilties, group_mid_points


    @torch.no_grad()
    def mutual_information_difference(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        sample_hypothesis_test: bool = False,
        samples: int = 1000,
        log_base: float = 2.0,
        **kwargs
    ) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        r"""
        Compute the mutual information difference (MID) and the absolute value of mutual information difference (AMID) statistic :cite:p:`Kim2011` for the provided data to test for conditional independence among items given :math:`\theta` (local independence). 

        Parameters
        ----------
        data : torch.Tensor
            The data used to compute the AMID statistic. Uses the model's training data if not provided.
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        sample_hypothesis_test : bool, optional
            Whether to sample from the null hypothesis distribution for the AMID statistic and perform a statistical test for each item pair. (default is False)
        samples : int, optional
            The number of samples to draw from the null hypothesis distribution. (default is 1000)
        log_base : float, optional
            The base of the logarithm used to compute the entropy. (default is 2.0)
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]
            A tuple with three data frames. The first two are the MID and AMID statistics for each item pair. The third data frame contains the p-values of the AMID tests if sample_hypothesis_test is True.

        Examples
        --------
        >>> import irtorch
        >>> from irtorch.models import MonotoneNN
        >>> from irtorch.estimation_algorithms import AE
        >>> data = irtorch.load_dataset.swedish_national_mathematics_1()
        >>> model = MonotoneNN(1, data)
        >>> model.fit(train_data=data, algorithm=AE())
        >>> mid, amid, p_value = model.evaluation.mutual_information_difference(data, sample_hypothesis_test=True, samples=300)
        """
        if data is None:
            data = self.model.algorithm.train_data
        else:
            data = data.contiguous()
        if theta is None:
            theta = self.model.latent_scores(data=data, rescale=False, **kwargs)

        data_joint_entropies = joint_entropy_matrix(data, log_base=log_base)
        data_entropies = data_joint_entropies.diag()
        data_entropy_sums = data_entropies.unsqueeze(dim=-1) + data_entropies.unsqueeze(dim=0)
        data_mutual_information = data_entropy_sums - data_joint_entropies
        # Note: data_mutual_information can be negative in rare cases when missing values are present since
        # it's computed using the observed values in each column. There may be no missing in an individal
        # column for the entropy but some values may still be ignored in the joint entropy if another column has missing.
        
        model_probabilities = self.model.probabilities_from_output(self.model(theta))
        # Compute the expected entropy for each item
        expected_item_entropies = entropy(model_probabilities.mean(dim=0), log_base=log_base)
        # Expected respondent proportions for each pairwise combination of item responses
        # Returns a 5D tensor with dimensions (Respondent, item 1, item 2, item 1 response, item 2 response)
        expected_proportions = torch.einsum('bik, bjl -> bijkl', model_probabilities, model_probabilities)
        # Average over respondents (integrate)
        expected_proportions = expected_proportions.mean(dim=0)

        expected_joint_entropies = entropy(
            expected_proportions.view(expected_proportions.shape[0], expected_proportions.shape[1], -1),
            log_base=log_base
        )
        expected_entropy_sums = expected_item_entropies.unsqueeze(dim=-1) + expected_item_entropies.unsqueeze(dim=0)
        expected_mutual_information = expected_entropy_sums - expected_joint_entropies

        mid = data_mutual_information - expected_mutual_information
        amid = torch.abs(mid)
        mid_df = pd.DataFrame(mid.detach().numpy())
        mid_df.columns = [f"Item {i+1}" for i in range(mid.shape[0])]
        mid_df.index = [f"Item {i+1}" for i in range(mid.shape[0])]
        amid_df = pd.DataFrame(amid.detach().numpy())
        amid_df.columns = [f"Item {i+1}" for i in range(mid.shape[0])]
        amid_df.index = [f"Item {i+1}" for i in range(mid.shape[0])]

        df_no_diag = amid_df.where(~np.eye(amid_df.shape[0], dtype=bool))
        max_value = df_no_diag.max().max()
        max_location = df_no_diag.stack().idxmax()
        logger.info("Largest AMID is %.2f between %s and %s.", max_value, max_location[0], max_location[1])

        if sample_hypothesis_test:
            sample_amids = torch.zeros(samples, *data_mutual_information.shape)
            for sample in range(samples):
                dynamic_print(f"Computing AMID p-values by null distribution sampling. Sample: {sample+1}")
                sample_data = self.model.sample_test_data(theta)
                sample_joint_entropies = joint_entropy_matrix(sample_data, log_base=log_base)
                sample_entropies = sample_joint_entropies.diag()
                sample_entropy_sums = sample_entropies.unsqueeze(dim=-1) + sample_entropies.unsqueeze(dim=0)
                sample_mutual_information = sample_entropy_sums - sample_joint_entropies
                sample_amids[sample, :, :] = (sample_mutual_information - expected_mutual_information).abs()

            observed_amid = amid.unsqueeze(0).expand(sample_amids.shape[0], -1, -1)
            # how many times are the samples larger than the observed amid?
            counts = (sample_amids > observed_amid).sum(dim=0)
            # if it is a relatively low amount of times, the p-value is low
            p_values = counts / sample_amids.shape[0]
            p_values.fill_diagonal_(torch.nan)

            p_values_df = pd.DataFrame(p_values.detach().numpy())
            p_values_df.columns = [f"Item {i+1}" for i in range(p_values.shape[0])]
            p_values_df.index = [f"Item {i+1}" for i in range(p_values.shape[0])]
            return mid_df, amid_df, p_values_df.round(3)
        
        return mid_df, amid_df, None


    @torch.no_grad()
    def sum_score_probabilities(
        self,
        latent_density_method: str = "data",
        population_data: torch.Tensor = None,
        trapezoidal_segments: int = 1000,
        sample_size: int = 100000,
    ):
        """
        Computes the marginal probabilities for each sum score, averged over the latent space density. For 'qmvn' and 'gmm' densities, the trapezoidal rule is used for integral approximation.

        Parameters
        ----------
        latent_density_method : str, optional
            Specifies the method used to approximate the latent space density.
            Possible options are
            - 'data' averages over the theta scores from the population data.
            - 'encoder sampling' samples theta scores from the encoder. Only available for VariationalAutoencoderIRT models
            - 'qmvn' for quantile multivariate normal approximation of a multivariate joint density function (QuantileMVNormal class).
            - 'gmm' for a gaussian mixture model.
        population_data : torch.Tensor, optional
            The population data used for approximating sum score probabilities. Default is None and uses the training data.
        trapezoidal_segments : int, optional
            The number of integration approximation intervals for each theta dimension. (Default is 1000)
        sample_size : int, optional
            Sample size for the 'encoder sampling' method. (Default is 100000)
        Returns
        -------
        torch.Tensor
            A 1D tensor with the probability for each total score.
        """
        self._validate_latent_density_method(latent_density_method)
        theta_scores, weights = self._get_theta_scores_and_weights(
            latent_density_method, population_data, trapezoidal_segments, sample_size
        )
        probabilities = self.model.item_probabilities(theta_scores)
        # sum incorrect response option probabilities if MC
        if self.model.mc_correct is not None:
            probabilities = sum_incorrect_probabilities(
                probabilities=probabilities,
                item_responses=self.model.item_categories,
                mc_correct=self.model.mc_correct
            )
            item_categories = [2] * len(self.model.item_categories)
        else:
            item_categories = self.model.item_categories
        
        conditional_total_score_probs = conditional_score_distribution(
            probabilities, item_categories
        )
        sum_score_probabilities = conditional_total_score_probs * weights.view(-1, 1)
        return sum_score_probabilities.sum(dim=0)

    def q3(
        self,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        sample_hypothesis_test: bool = False,
        samples: int = 1000,
        **kwargs
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        r"""
        Compute the Q3 statistic :cite:p:`Kim2011` for the provided data to test for conditional independence among items given :math:`\theta` (local independence).

        Parameters
        ----------
        data : torch.Tensor
            The data used to compute the Q3 statistic. Uses the model's training data if not provided.
        theta: torch.Tensor, optional
            The latent variable theta scores for the provided data on the original theta scale.
            If not provided, they will be computed using :meth:`irtorch.models.BaseIRTModel.latent_scores`.
        sample_hypothesis_test : bool, optional
            Whether to sample from the null hypothesis distribution for the Q3 statistic and perform a statistical test for each item pair. (default is False)
        samples : int, optional
            The number of samples to draw from the null hypothesis distribution. (default is 1000)
        **kwargs : dict, optional
            Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            A tuple with the Q3 statistic for each item pair and the corresponding p-values of the Q3 tests if sample_hypothesis_test is True.

        Examples
        --------
        >>> from irtorch.models import MonotoneNN
        >>> from irtorch.estimation_algorithms import AE
        >>> from irtorch.load_dataset import swedish_national_mathematics_1
        >>> data = swedish_national_mathematics_1()
        >>> model = MonotoneNN(1, data)
        >>> model.fit(train_data=data, algorithm=AE())
        >>> q3, p_value = model.evaluation.q3(data, sample_hypothesis_test=True, samples=300)
        """
        if data is None:
            data = self.model.algorithm.train_data
        else:
            data = data.contiguous()
        if theta is None:
            theta = self.model.latent_scores(data=data, rescale=False, **kwargs)

        residuals = self.residuals(data=data, theta=theta, average_over="none", **kwargs)
        corr_matrix = correlation_matrix(residuals)
        corr_matrix.fill_diagonal_(torch.nan)
        corr_matrix_df = pd.DataFrame(corr_matrix.detach().numpy())
        corr_matrix_df.columns = [f"Item {i+1}" for i in range(corr_matrix.shape[0])]
        corr_matrix_df.index = [f"Item {i+1}" for i in range(corr_matrix.shape[0])]

        # Find the maximum value and its location
        df_no_diag = corr_matrix_df.where(~np.eye(corr_matrix_df.shape[0], dtype=bool))
        max_value = df_no_diag.max().max()
        max_location = df_no_diag.stack().idxmax()
        logger.info("Largest Q3 is %.2f between %s and %s.", max_value, max_location[0], max_location[1])

        if sample_hypothesis_test:
            sample_corr_matrices = torch.zeros(samples, *corr_matrix.shape)
            for sample in range(samples):
                dynamic_print(f"Computing Q3 p-values by null distribution sampling. Sample: {sample+1}")
                sample_data = self.model.sample_test_data(theta)
                sample_residuals = self.residuals(
                    sample_data,
                    theta = theta,
                    average_over="none",
                    **kwargs
                )
                sample_corr_matrices[sample, :, :] = correlation_matrix(sample_residuals)

            observed_q3 = corr_matrix.unsqueeze(0).expand(sample_corr_matrices.shape[0], -1, -1)
            counts1 = (sample_corr_matrices < observed_q3).sum(dim=0) # how many samples are smaller than the obs. correlation?
            counts2 = (sample_corr_matrices >= observed_q3).sum(dim=0) # how many samples are larger than the obs. correlation?
            # if most samples are smaller than the observed correlation, the p-value is the proportion of samples that are larger
            # times 2 since we are interested in both tails
            p_values = torch.min(counts1, counts2).float() * 2 / sample_corr_matrices.shape[0]
            p_values.fill_diagonal_(torch.nan)

            p_values_df = pd.DataFrame(p_values.detach().numpy())
            p_values_df.columns = [f"Item {i+1}" for i in range(corr_matrix.shape[0])]
            p_values_df.index = [f"Item {i+1}" for i in range(corr_matrix.shape[0])]
            return corr_matrix_df.round(3), p_values_df.round(3)
        
        return corr_matrix_df.round(3), None

    # def _expected_item_score_combination_proportions(self, response_probabilities: torch.Tensor) -> torch.Tensor:
    #     """
    #     Computes the item score proportions for each pairwise combination of item responses, averaged over the latent space.

    #     Parameters
    #     ----------
    #     response_probabilities : torch.Tensor
    #         A 3D tensor with the probabilities for each item response. The first dimension represents the respondents, the second dimension represents the items and the third dimension represents the item categories.

    #     Returns
    #     -------
    #     torch.Tensor
    #         A 4D tensor with the expected item score proportions. The first two dimensions represent the item pair, and the last two dimensions represent the item categories.
    #     """
    #     expected_combos = torch.zeros(
    #         response_probabilities.shape[0],
    #         response_probabilities.shape[1],
    #         response_probabilities.shape[1],
    #         response_probabilities.shape[2],
    #         response_probabilities.shape[2],
    #     )
    #     for i in range(response_probabilities.shape[1]):
    #         for j in range(response_probabilities.shape[1]):
    #             for k in range(response_probabilities.shape[2]):
    #                 for l in range(response_probabilities.shape[2]):
    #                     expected_combos[:, i, j, k, l] = response_probabilities[:, i, k] * response_probabilities[:, j, l]

    #     return expected_combos.mean(dim=0)

    def _observed_item_score_proportions(self, data: torch.Tensor) -> torch.Tensor:
        """
        Computes the observed item score proportions for each item in the data.

        Parameters
        ----------
        data : torch.Tensor
            The input data.

        Returns
        -------
        torch.Tensor
            A 2D tensor with the observed item score proportions. The first dimension represents the items and the second dimension represents the item categories.
        """
        item_responses = self.model.item_categories
        item_score_proportions = torch.zeros(len(item_responses), max(item_responses))
        missing_mask = get_missing_mask(data)
        for item_i, _ in enumerate(item_responses):
            counts = torch.bincount(data[:, item_i][~missing_mask[:, item_i]].int(), minlength=max(item_responses))
            item_score_proportions[item_i, :] = counts.float() / counts.sum()

        return item_score_proportions

    @torch.no_grad()
    def _min_max_theta_for_integration(
        self,
        theta: torch.Tensor = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Retrieve the minimum and maximum theta score for approximating integrals over the latent space. Uses one standard deviation below/above the min/max of each theta score vector.

        Parameters
        ----------
        theta : torch.Tensor, optional
            A 2D tensor. Columns are each latent variable, rows are respondents. Default is None and uses training data theta scores.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor]
            A tuple with 1D tensors, containing the min and max integration theta scores of each latent variable.
        """
        if theta is None:
            if self.model.algorithm.training_theta_scores is not None:
                theta = self.model.algorithm.training_theta_scores
                theta_min = theta.min(dim=0)[0]
                theta_max = theta.max(dim=0)[0]
                theta_stds = theta.std(dim=0)
            else:
                theta_min = torch.full((self.model.latent_variables,), -3)
                theta_max = torch.full((self.model.latent_variables,), 3)
                theta_stds = torch.ones(self.model.latent_variables)
        else:
            theta_min = theta.min(dim=0)[0]
            theta_max = theta.max(dim=0)[0]
            theta_stds = theta.std(dim=0)

        return theta_min - theta_stds, theta_max + theta_stds

    @torch.no_grad()
    def _grouped_theta_probabilities(self, grouped_theta: tuple[torch.Tensor, ...]):
        """
        Computes the average probabilities for each potential item response for each group.

        Parameters
        ----------
        grouped_theta : tuple[torch.Tensor, ...]
            A tuple containing 2D tensors. Each tensor represents a group of respondents, with the first dimension corresponding to the respondents and the second dimension representing their latent variables in the form of theta-scores.

        Returns
        -------
        torch.Tensor
            A 3D torch tensor with group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.
        """
        group_probabilities = torch.zeros(len(grouped_theta), len(self.model.item_categories), max(self.model.item_categories))
        for group_i, group in enumerate(grouped_theta):
            item_probabilities = self.model.item_probabilities(group)
            group_probabilities[group_i, :, :] = item_probabilities.mean(dim=0)
        return group_probabilities

    @torch.no_grad()
    def _grouped_data_probabilities(self, grouped_data: tuple[torch.Tensor, ...]):
        """
        Computes the average probabilities for each potential item response for each group.

        Parameters
        ----------
        grouped_data : tuple[torch.Tensor, ...]
            A tuple containing 2D tensors. Each tensor represents a group of respondents, with the first dimension representing the respondents and the second dimension representing items.

        Returns
        -------
        torch.Tensor
            A 3D torch tensor with group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.
        """
        item_responses = self.model.item_categories
        group_probabilities = torch.zeros(len(grouped_data), len(item_responses), max(item_responses))
        for group_i, group in enumerate(grouped_data):
            missing_mask = get_missing_mask(group)
            for item_i, _ in enumerate(item_responses):
                counts = torch.bincount(group[:, item_i][~missing_mask[:, item_i]].int(), minlength=max(item_responses))
                group_probabilities[group_i, item_i, :] = counts.float() / counts.sum()

        return group_probabilities

    def _validate_latent_density_method(self, latent_density_method: str) -> None:
        valid_methods = ["data", "encoder sampling", "qmvn", "gmm"]
        if latent_density_method not in valid_methods:
            raise ValueError(
                f"Invalid latent density method. Must be one of {valid_methods}."
            )
        if latent_density_method == "encoder sampling" and not isinstance(
            self.model.algorithm, VAE
        ):
            raise ValueError(
                "Encoder sampling is only available for variational autoencoder models."
            )

    def _get_theta_scores_and_weights(
        self,
        latent_density_method: str,
        population_data: torch.Tensor,
        trapezoidal_segments: int,
        sample_size: int,
        ml_map_device: str = "cuda" if torch.cuda.is_available() else "cpu",
        lbfgs_learning_rate: float = 0.25,
    ):
        if population_data is None:
            if self.model.algorithm.training_theta_scores is not None:
                theta_scores = self.model.algorithm.training_theta_scores
            elif isinstance(self.model.algorithm, MML):
                logger.info("Sampling from multivariate normal as population theta scores.")
                mvn = MultivariateNormal(torch.zeros(self.model.latent_variables), self.model.algorithm.covariance_matrix)
                theta_scores = mvn.sample((4000,)).to(dtype=torch.float32)
        else:
            theta_scores = self.model.latent_scores(population_data, theta_estimation="NN", device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate, rescale=False)

        if latent_density_method in ["data", "encoder sampling"]:
            weights = (
                torch.ones(theta_scores.shape[0])
                / theta_scores.shape[0]
            )
            if latent_density_method == "encoder sampling":
                theta_scores = self.model.algorithm.sample_latent_variables(
                    sample_size=sample_size, input_data=population_data
                )
        else:
            (
                theta_scores,
                weights,
            ) = self._get_theta_scores_and_weights_for_latent_density_methods(
                latent_density_method, theta_scores, population_data, trapezoidal_segments
            )

        return theta_scores, weights

    def _get_theta_scores_and_weights_for_latent_density_methods(
        self,
        latent_density_method: str,
        theta_scores: torch.Tensor,
        population_data: torch.Tensor,
        trapezoidal_segments: int,
    ):
        # We approximate the density if
        # population_data is not none and the correct density
        # is not already in self.model.evaluation.latent_density
        if population_data is not None or (
            (
                latent_density_method != "qmvn"
                or not isinstance(self.latent_density, QuantileMVNormal)
            )
            and (
                latent_density_method != "gmm"
                or not isinstance(self.latent_density, GaussianMixtureModel)
            )
        ):
            self.approximate_latent_density(
                theta_scores=theta_scores, approximation=latent_density_method
            )

        # get the min/max points for integration
        min_theta, max_theta = self._min_max_theta_for_integration(theta_scores)

        # Create a list of linspace tensors for each dimension
        lin_spaces = [
            torch.linspace(min_theta[i], max_theta[i], trapezoidal_segments)
            for i in range(len(min_theta))
        ]

        # Use torch.cartesian_prod to generate all combinations
        theta_scores = torch.cartesian_prod(*lin_spaces)

        if theta_scores.dim() == 1:
            # Add an extra dimension for 1D models to make it a 2D tensor with 1 column
            theta_scores = theta_scores.unsqueeze(1)

        weights = self.latent_density.pdf(theta_scores)
        weights = weights / weights.sum()

        return theta_scores, weights


from .base_irt_model import BaseIRTModel
from .monotone_nn import MonotoneNN
from .model_mix import ModelMix
from .one_parameter_logistic import OneParameterLogistic
from .two_parameter_logistic import TwoParameterLogistic
from .three_parameter_logistic import ThreeParameterLogistic
from .generalized_partial_credit import GeneralizedPartialCredit
from .graded_response import GradedResponse
from .nominal_response import NominalResponse
from .monotone_polynomial import MonotonePolynomial



import logging
import copy
import torch
import numpy as np
from irtorch.models import BaseIRTModel
from irtorch.estimation_algorithms import BaseIRTAlgorithm
from irtorch.estimation_algorithms.encoders import StandardEncoder
from irtorch._internal_utils import dynamic_print
from irtorch.irt_dataset import PytorchIRTDataset

logger = logging.getLogger("irtorch")

class AE(BaseIRTAlgorithm):
    """
    Autoencoder neural network for fitting IRT models.
    """
    def __init__(self):
        super().__init__()
        self.one_hot_encoded = True
        self.batch_normalization = False
        self.encoder = None
        self.imputation_method = None
        self.data_loader = None
        self.optimizer = None
        self.item_categories = None
        self.mc_correct = None
        self.training_history = {
            "train_loss": [],
        }
        self.best_model_state = None
        self.batch_mean_losses = []
        self.best_avg_loss = float("Inf"), 0
        self.total_iterations = 0
        self.lr_update_count = 0
        self.evaluation_interval_size = 60

    def fit(
        self,
        model: BaseIRTModel,
        train_data: torch.Tensor,
        one_hot_encoded: bool = True,
        imputation_method: str = None,
        learning_rate: float = 0.02,
        learning_rate_updates_before_stopping: int = 3,
        evaluation_interval_size: int = 80,
        max_epochs: int = 10000,
        batch_size: int = None,
        batch_normalization_encoder: bool = False,
        nonlinear_encoder = torch.nn.ELU(),
        hidden_layers_encoder: list[int] = None,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
    ):
        """
        Train an IRT model using the autoencoder. If the algorithm fails to converge, try lowering the learning rate. Use batch_size if the data is too large to fit in memory.

        Parameters
        ----------
        model : BaseIRTModel
            The model to fit. Needs to inherit :class:`irtorch.models.BaseIRTModel`.
        train_data : torch.Tensor
            The training data. Item responses should be coded 0, 1, ... and missing responses coded as nan or -1.
        one_hot_encoded : bool, optional
            Whether or not to one-hot encode the train data as encoder input inside this fit method. (default is True)
        imputation_method : str, optional
            The method to use for imputing missing data for the encoder. For options see :func:`irtorch.utils.impute_missing`.
            Only methods not relying on a fitted model can be used. 
            Note that missing values are removed from the loss calculation even after imputation.
            If you do not want this, do the imputation to your dataset before fitting. (default is None and only works for one-hot encoded inputs)
        learning_rate : float, optional
            The initial learning rate for the optimizer. (default is 0.02)
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training. (default is 3)
        evaluation_interval_size: int, optional
            The number of iterations between each model evaluation during training. (default is 80)
        max_epochs : int, optional
            The maximum number of epochs to train for. (default is 1000)
        batch_size : int, optional
            The batch size for training. (default is None and uses the full dataset)
        batch_normalization_encoder : bool, optional
            Whether to use batch normalization for the encoder. (default is True)
        nonlinear_encoder : torch.nn.Module, optional
            The non-linear function to use after each hidden layer in the encoder. (default is torch.nn.ELU())
        hidden_layers_encoder : list[int], optional
            List of hidden layers for the encoder. Each element is a layer with the number of neurons represented as integers. If not provided, uses one hidden layer with 2 * sum(item_categories) neurons.
        device : str, optional
            The device to run the model on. (default is "cuda" if available else "cpu".)
        """
        super().fit(model = model, train_data = train_data)
        # Re-initialize the training history
        self.training_history = {
            "train_loss": [],
        }
        self.best_model_state = None
        self.batch_mean_losses = []
        self.best_avg_loss = float("Inf"), 0
        self.total_iterations = 0
        self.lr_update_count = 0

        self.one_hot_encoded = one_hot_encoded
        self.batch_normalization = batch_normalization_encoder
        self.imputation_method = imputation_method
        self.item_categories = model.item_categories
        self.mc_correct = model.mc_correct
        self.evaluation_interval_size = evaluation_interval_size

        if batch_size is None:
            batch_size = train_data.shape[0]
        if self.one_hot_encoded:
            input_dim = sum(model.item_categories)
        else:
            input_dim = len(model.item_categories)

        if hidden_layers_encoder is None:  # 1 layer with 2x number of categories as neurons is default
            hidden_layers_encoder = [2 * sum(model.item_categories)]

        self.encoder = StandardEncoder(
            input_dim=input_dim,
            latent_variables=model.latent_variables,
            hidden_dim=hidden_layers_encoder,
            batch_normalization=batch_normalization_encoder,
            nonlinear=nonlinear_encoder,
        )

        if not one_hot_encoded and imputation_method is None:
            raise ValueError("imputation_method must be supplied for non-one-hot encoded data.")

        train_data_irt = PytorchIRTDataset(
            data=train_data.to(device),
            one_hot_encoded=self.one_hot_encoded,
            item_categories=model.item_categories,
            imputation_method=imputation_method,
            mc_correct=model.mc_correct
        )
        self.data_loader = torch.utils.data.DataLoader(
            train_data_irt,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=False,
        )

        self.optimizer = torch.optim.Adam(
            list(self.encoder.parameters()) + list(model.parameters()), lr=learning_rate, amsgrad=True
        )

        # Reduce learning rate when loss stops decreasing ("min")
        # we multiply the learning rate by the factor
        # patience: We need no improvement every step for it to trigger
        self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, mode="min", factor=0.6, patience=1
        )

        self.encoder.to(device)
        model.to(device)
        self._training_loop(model, max_epochs, learning_rate_updates_before_stopping)
        self.encoder.to("cpu")
        model.to("cpu")
        self.encoder.eval()
        model.eval()

        # store the latent theta scores of the training data
        # used for more efficient computation when using other methods
        self.training_theta_scores = self.theta_scores(
            train_data_irt.input_data.to("cpu")
        ).clone().detach()

    def _training_loop(
        self,
        model: BaseIRTModel,
        max_epochs: int,
        learning_rate_updates_before_stopping: int = 2,
    ):
        """
        The training loop for the model.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        max_epochs : int
            The maximum number of epochs to train for.
        scheduler : torch.optim.lr_scheduler.ReduceLROnPlateau
            The learning rate scheduler.
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training. (default is 2)
        """
        for epoch in range(max_epochs):
            train_loss = self._train_step(model, epoch, learning_rate_updates_before_stopping)
            if np.isnan(train_loss):
                break
            if self.lr_update_count >= learning_rate_updates_before_stopping:
                break

        # Load the best model state
        if self.best_model_state is not None:
            model.load_state_dict(self.best_model_state["state_dict_model"])
            self.encoder.load_state_dict(self.best_model_state["state_dict_encoder"])
            self.optimizer.load_state_dict(self.best_model_state["optimizer"])
            logger.info("Best model found after %d iterations (%d interval updates) with interval averaged loss %.4f.", self.best_avg_loss[1], self.best_avg_loss[1]/self.evaluation_interval_size, self.best_avg_loss[0])

    def _train_step(self, model: BaseIRTModel, epoch: int, learning_rate_updates_before_stopping: int = 5) -> torch.Tensor:
        """
        Perform a training step for the model.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        epoch : int
            The current epoch.
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training. (default is 5)

        Returns
        -------
        float
            The loss after the training step.
        """
        self.encoder.train()
        model.train()
        epoch_mean_loss = 0
        learning_rate = self.optimizer.param_groups[0]['lr']
        for _, (batch, mask, input_batch, _) in enumerate(self.data_loader):
            # small batches leads to inaccurate batch variance, so we drop the last few observations
            if batch.shape[0] < 4 and self.batch_normalization:
                continue
            self.optimizer.zero_grad()
            batch_loss = self._train_batch(model, input_batch, batch, mask)
            if torch.isnan(batch_loss):
                logger.warning("Batch loss is nan. Try increasing batch size or lowering the learning rate.")
                return float("nan")

            batch_loss.backward()

            self.optimizer.step()
            self.total_iterations += 1
            epoch_mean_loss += batch_loss.item()

            avg_batch_loss = batch_loss.item()
            dynamic_print(f"Epoch: {epoch + 1}. Iteration: {self.total_iterations}. Average batch loss: {avg_batch_loss:.4f}. Current learning rate: {learning_rate:.4f}")

            self.batch_mean_losses.append(avg_batch_loss)
            if len(self.batch_mean_losses) > self.evaluation_interval_size:
                self.batch_mean_losses.pop(0)

            # Update the learning rate scheduler every self.evaluation_interval_size iterations if no improvement
            if (self.total_iterations) % self.evaluation_interval_size == 0 and self.total_iterations != 1:
                mean_loss = np.mean(self.batch_mean_losses)
                self.training_history["train_loss"].append(mean_loss)
                self.scheduler.step(mean_loss)

                # Check if the learning rate has been updated
                if learning_rate != self.optimizer.param_groups[0]['lr']:
                    self.lr_update_count += 1
                    learning_rate = self.optimizer.param_groups[0]['lr']
                
                if mean_loss < self.best_avg_loss[0]:
                    self.best_avg_loss = mean_loss, self.total_iterations
                    self.best_model_state = {
                        "state_dict_model": copy.deepcopy(model.state_dict()),
                        "state_dict_encoder": copy.deepcopy(self.encoder.state_dict()),
                        "optimizer": copy.deepcopy(self.optimizer.state_dict()) 
                    }
                
                if self.lr_update_count >= learning_rate_updates_before_stopping:
                    logger.info("Stopping training after %s learning rate updates.", learning_rate_updates_before_stopping)
                    break

        epoch_mean_loss /= len(self.data_loader)
        return epoch_mean_loss

    def _train_batch(self, model: BaseIRTModel, input_batch: torch.Tensor, batch: torch.Tensor, missing_mask: torch.Tensor):
        """
        Train the model on a batch of data.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        input_batch : torch.Tensor
            The input batch for the encoder.
        batch : torch.Tensor
            The batch of data for computing the loss.
        missing_mask : torch.Tensor
            The mask for missing data.
            
        Returns
        -------
        tuple
            The logits and loss after training on the batch.
        """
        outputs = model(self.encoder(input_batch))
        # check if all outputs are nan
        if torch.isnan(outputs).all():
            return torch.tensor(torch.nan)
        batch_loss = -model.log_likelihood(batch, outputs, missing_mask) / batch.shape[0]
        return batch_loss

    @torch.no_grad()
    def _batch_fit_measures(self, model: BaseIRTModel, input_batch: torch.Tensor, batch: torch.Tensor, missing_mask: torch.Tensor):
        """
        Calculate the fit measures for a batch.

        Parameters
        ----------
        model : BaseIRTModel
            The model to calculate the fit measures for.
        input_batch : torch.Tensor
            The input batch for the encoder.
        batch : torch.Tensor
            The batch of data.
        missing_mask : torch.Tensor
            The mask for missing data.

        Returns
        -------
        tuple
            The loss, log likelihood, and accuracy for the batch.
        """
        output = model(self.encoder(input_batch))
        log_likelihood = model.log_likelihood(batch, output, missing_mask=missing_mask)
        loss = -log_likelihood / batch.shape[0]
        return loss

    @torch.no_grad()
    def theta_scores(self, data: torch.Tensor):
        """
        Get the latent scores from an input dataset using the encoder.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents.

        Returns
        -------
        torch.Tensor
            A 2D tensor of latent scores. Rows are respondents and latent variables are columns.
        """
        data = data.contiguous()
        return self.encoder(data)



import logging
import torch
from irtorch.rescale import Scale

logger = logging.getLogger("irtorch")

class Reverse(Scale):
    """
    Reverses the chosen theta scales using.
    
    Parameters
    ----------
    reversed_latent_variables : list[bool]
        A list of booleans indicating which latent variables to reverse.
    
    Examples
    --------
    >>> import irtorch
    >>> from irtorch.models import NominalResponse, MonotoneNN
    >>> from irtorch.estimation_algorithms import AE, MML
    >>> from irtorch.rescale import Reverse
    >>> irtorch.set_seed(15)
    >>> data_sat, correct_responses = irtorch.load_dataset.swedish_sat_verbal()
    >>> model = NominalResponse(data=data_sat, mc_correct=correct_responses)
    >>> model.fit(train_data=data_sat, algorithm=MML())
    >>> model.plot.plot_item_probabilities(1).show()
    >>> # reverse the first (and only) latent variable
    >>> reverse = Reverse([True])
    >>> model.add_scale_tranformation(reverse)
    >>> model.plot.plot_item_probabilities(1).show()
    """
    def __init__(self, reversed_latent_variables: list[bool]):
        super().__init__(invertible=True)
        self._reverse = -torch.tensor(reversed_latent_variables, dtype=int).reshape(1, -1)
        self._reverse[self._reverse == 0] = 1

    def transform(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Transforms the input theta scores into the new scale.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.
        """
        
        return theta * self._reverse

    def inverse(self, transformed_theta: torch.Tensor) -> torch.Tensor:
        """
        Puts the scores back to the original theta scale.

        Parameters
        ----------
        transformed_theta : torch.Tensor
            A 2D tensor containing transformed theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A 2D tensor containing theta scores on the the original scale.
        """
        return transformed_theta * self._reverse

    def jacobian(
        self,
        theta: torch.Tensor
    ) -> torch.Tensor:
        r"""
        Computes the gradients of scale scores for each :math:`j` with respect to the input theta scores.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A torch tensor with the gradients for each theta score. Dimensions are (theta rows, latent variables, latent variables) where the last two are the jacobians.
        """
        return torch.diag_embed(self._reverse.flatten().float()).unsqueeze(0).expand(theta.shape[0], -1, -1)



import torch
from irtorch._internal_utils import one_hot_encode_test_data
from irtorch._internal_utils import impute_missing_internal

class PytorchIRTDataset(torch.utils.data.Dataset):
    """
    PyTorch dataset for IRT data.

    Parameters
    ----------
    data : torch.Tensor
        A 2D tensor with the data. Rows are respondents and columns are items.
    one_hot_encoded : bool, optional
        Only relevant for autoencoder models. Whether the input data is one-hot encoded. (default is False)
    item_categories : list[int], optional
        For one_hot_encoding. A list of integers where each integer is the number of possible responses for the corresponding item. (default is None)
    imputation_method : str, optional
        The method to use for imputing missing values into the autoencoder input data.
        The data itself is not modified. 
        See :func:`irtorch.utils.impute_missing` for available methods.
        Note that only methods not relying on a fitted model can be used. (The default is None)
    **kwargs
        Additional keyword arguments passed to :func:`irtorch.utils.impute_missing`.
    """
    def __init__(self,
        data: torch.Tensor,
        one_hot_encoded: bool = False,
        item_categories: list[int] = None,
        imputation_method: str = None,
        **kwargs
    ):
        super().__init__()
        self.data = data.clone() # clone to avoid modifying the original data

        # set missing responses to 0 in the response mask (all non-missing are ones)
        self.mask = (data == -1) | data.isnan()

        if imputation_method is not None:
            data = impute_missing_internal(data = data, method=imputation_method, item_categories=item_categories, **kwargs)
        
        if one_hot_encoded:
            if item_categories is None:
                raise ValueError("item_categories must be supplied for one-hot encoded data.")
            self.input_data = one_hot_encode_test_data(data, item_categories)
        else:
            self.input_data = data

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, index: int):
        """
        Get a sample from the dataset.

        Parameters
        ----------
        index : int
            The index of the sample to get.

        Returns
        -------
        torch.Tensor
            The data for the sample.
        torch.Tensor
            The mask for the sample.
        torch.Tensor
            Only relevant for autoencoder models. The autoencoder input data for the sample.
        """
        return self.data[index], self.mask[index], self.input_data[index], index


import logging
from .evaluator import Evaluator
from .plotter import Plotter
from . import load_dataset
from .config import *
from .utils import *

__version__ = "0.4.1"

logging.basicConfig(level=logging.INFO,
                    format="%(levelname)s: %(message)s",
                    datefmt="%Y-%m-%d %H:%M:%S")

logger = logging.getLogger("irtorch")
